################################################################################
# Molecule Class                                                               #
#                                                                              #
"""All necessary function for creating and editing molecules."""
################################################################################


import math

import porems.utils as utils


class Molecule:
    """This class defines a molecule object, which is basically a data matrix
    that contains atom information of positions and type.

    Be Molecule :math:`\\boldsymbol{D}\\in\\mathbb{R}^{(n+1)\\times m}` with spatial dimensions
    :math:`i=1,\\dots,n` and atom number :math:`j=1,\\dots,m`.
    The data matrix then has following structure

    .. math::

        \\boldsymbol{D}=\\begin{bmatrix}
        \\begin{pmatrix}d_{11}\\\\d_{12}\\\\\\vdots\\\\d_{1m}\\end{pmatrix}&
        \\begin{pmatrix}d_{21}\\\\d_{22}\\\\\\vdots\\\\d_{2m}\\end{pmatrix}&
        \\dots&
        \\begin{pmatrix}d_{n1}\\\\d_{n2}\\\\\\vdots\\\\d_{nm}\\end{pmatrix}&
        \\begin{pmatrix}d_{t1}\\\\d_{t2}\\\\\\vdots\\\\d_{tm}\\end{pmatrix}
        \\end{bmatrix}

    with atom data :math:`d_{ij}` of atom :math:`j` and chemical atom type entry :math:`t:=n+1`.

    The requested dimension is defined in the parameter module.
    Most of the following functions have been written for :math:`n`-dimensions.
    However they only have been tested for the three-dimnesional case.

    Functions have been provided for editing, moving and transforming the
    entries of the data matrix either for the whole matrix or for specified atoms.
    Private functions are solely needed for internal puposes,
    whereas the public functions are inteded to be used for the editingself.

    Parameters
    ----------
    name : string, optional
        Molecule name
    short : string, optional
        Molecule short name
    inp : None, string, list, optional
        None for empty Molecule, string to read a molecule from a
        specified filelink or a list of either molecules to concatenate these
        into one object, or a data list in the earlier discussed format
    """
    def __init__(self, name="Molecule", short="MOL", inp=None):
        # Initialize
        self._box = None
        self._masses = None
        self._mass = None
        self._com = None
        self._mol_list = [self]

        self._dim = 3
        self._charge = 0

        self._name = name
        self._short = short

        # Check data input
        if inp is None:
            self._data = [[] for i in range(self._dim+1)]
        else:
            # Read from file
            if isinstance(inp, str):
                self._data = self._read(inp, inp.split(".")[-1].upper())
            # Concat multiple molecules
            elif isinstance(inp, list):
                # Data list is provided
                if(isinstance(inp[0], list)):
                    self._data = inp
                # List of molecules is provided
                else:
                    self._data = self._concat(inp)


    ################################
    # Private Methods - Management #
    ################################
    def _read(self, file_path, file_type):
        """Read a molecule from a file. Currently only **GRO**, **PFB** and
        **MOL2** files are supported. In case the ``file_type`` is **OBJ**, the
        function will attempt unpickling the object file. Note that laoding the
        object this way, only the data will be imported.

        Parameters
        ----------
        file_path : string
            Link to requested file
        file_type : string
            Int for types list id or str for file extension name

        Returns
        -------
        data : list
            data matrix
        """
        # Process input
        if not file_type in ["GRO", "PDB", "MOL2", "OBJ"]:
            print("Unsupported filetype.")
            return

        # Load object
        if file_type == "OBJ":
            try:
                return utils.load(file_path).get_data()
            except:
                print("Erroneous object file.")
                return

        # Read molecule
        data = []
        with open(file_path, "r") as file_in:
            line_idx = 0
            for line in file_in:
                line_val = line.split()

                # Gro file
                if file_type == "GRO":
                    if line_idx > 0 and len(line_val) > 3:
                        coord = [float(line_val[i]) for i in range(3, 5+1)]
                        coord.append(''.join([i for i in line_val[1] if not i.isdigit()]))
                        data.append(coord)

                # Pdb file
                elif file_type == "PDB":
                    if line_val[0] in ["ATOM", "HETATM"]:
                        coord = [float(line_val[i])/10 for i in range(6, 8+1)]
                        coord.append(line_val[11])
                        data.append(coord)

                # Mol2 file
                elif file_type == "MOL2":
                    if len(line_val) > 8:
                        coord = [float(line_val[i])/10 for i in range(2, 4+1)]
                        coord.append(''.join([i for i in line_val[1] if not i.isdigit()]))
                        data.append(coord)

                # Running variable
                line_idx += 1

        # Transform to column
        return utils.column(data)

    def _concat(self, mol_list):
        """Concatenate a molecule list into one molecule object.

        Parameters
        ----------
        mol_list : list
            List of molecule objects to be concatenated

        Returns
        -------
        data : list
            data matrix
        """
        # Initialize
        data = [[] for x in range(self._dim+1)]

        # Concatenate
        for mol in mol_list:
            # Molecules
            mol_data = mol.get_data()
            for i in range(len(mol_data)):
                for j in range(len(mol_data[i])):
                    data[i].append(mol_data[i][j])

        return data

    def _temp(self, atoms):
        """Create a temporary molecule of specified atom ids.

        Parameters
        ----------
        atoms : list
            List of atoms to be duplicated

        Returns
        -------
        mol : Molecule
            Molecule object
        """
        # Create temp matrix
        temp = [[] for x in range(self._dim+1)]

        # Fill matrix
        for i in range(len(atoms)):
            for j in range(self._dim+1):
                temp[j].append(self._data[j][atoms[i]])

        # Create molecule
        return Molecule(inp=temp)

    def _update(self, data, atoms):
        """Exchange data of specified atom with given data.

        Parameters
        ----------
        data : list
            New data list
        atoms : list
            List of atoms to be updated
        """
        for i in range(len(atoms)):
            for j in range(self._dim):
                self._data[j][atoms[i]] = data[j][i]

    def _append(self, mol):
        """Append a given molecule to the current object.

        Parameters
        ----------
        mol : Molecule
            Molecule object
        """
        temp = mol.get_data()
        for i in range(len(temp)):
            for j in range(len(temp[i])):
                self._data[i].append(temp[i][j])


    ##############################
    # Private Methods - Geometry #
    ##############################
    def _dotproduct(self, vec_a, vec_b):
        """Calculate the dotproduct of two vectors
        :math:`\\boldsymbol{a},\\boldsymbol{b}\\in\\mathbb{R}^n`

        .. math::

            \\text{dot}(\\boldsymbol{a},\\boldsymbol{b})=
            \\begin{pmatrix}a_1\\\\\\vdots\\\\a_n\\end{pmatrix}\\cdot
            \\begin{pmatrix}b_1\\\\\\vdots\\\\b_n\\end{pmatrix}=
            a_1\\cdot b_1+a_2\\cdot b_2+\\dots+a_n\\cdot b_n.

        Parameters
        ----------
        vec_a : list
            First vector :math:`\\boldsymbol{a}`
        vec_b : list
            Second vector :math:`\\boldsymbol{b}`

        Returns
        -------
        dot : float
            Dotproduct value
        """
        return sum((a*b) for a, b in zip(vec_a, vec_b))

    def _length(self, vec):
        """Calculate the length of a vector :math:`\\boldsymbol{a}\\in\\mathbb{R}^n`

        .. math::

            \\text{length}(\\boldsymbol{a})=|\\boldsymbol{a}|=\\sqrt{\\boldsymbol{a}\cdot\\boldsymbol{a}}

        Parameters
        ----------
        vec : list
            Vector a

        Returns
        -------
        length : float
            Vector length
        """
        return math.sqrt(self._dotproduct(vec, vec))

    def _vector(self, pos_a, pos_b):
        """Calculate the vector between to two positions
        :math:`\\boldsymbol{a},\\boldsymbol{b}\\in\\mathbb{R}^n`

        .. math::

            \\text{vec}(\\boldsymbol{a},\\boldsymbol{b})=\\begin{pmatrix}b_1-a_1\\\\\\vdots\\\\b_n-a_n\\end{pmatrix}

        The two inputs can either be atom indices or to vectorial positions.

        Parameters
        ----------
        pos_a : integer, list
            First position :math:`\\boldsymbol{a}`
        pos_b : integer, list
            Second position :math:`\\boldsymbol{b}`

        Returns
        -------
        vector : list
            Bond vector
        """
        # Process input
        if isinstance(pos_a, int) and isinstance(pos_b, int):
            pos_a = self.pos(pos_a)
            pos_b = self.pos(pos_b)
        elif not isinstance(pos_a, list) and isinstance(pos_b, list):
            print("Vector: Wrong input...")
            return

        # Check dimensions
        if not len(pos_a) == len(pos_b) or not len(pos_a) == self._dim:
            print("Vector: Wrong dimensions...")
            return

        # Calculate vector
        return [pos_b[i]-pos_a[i] for i in range(self._dim)]

    def _unit(self, vec):
        """Transform a vector :math:`\\boldsymbol{a}\\in\\mathbb{R}^n` into a unit vactor

        .. math::

            \\text{unit}(\\boldsymbol{a})=\\frac{\\boldsymbol{a}}{|\\boldsymbol{a}|}

        Parameters
        ----------
        vec : list
            Vector a

        Returns
        -------
        vec : list
            Vector
        """
        length = self._length(vec)

        return [vec[i]/length if not length == 0 else vec[i] for i in range(self._dim)]

    def _cross(self, vec_a, vec_b):
        """Calculate the crossproduct of two three-dimensional vectors
        :math:`\\boldsymbol{a},\\boldsymbol{b}\\in\\mathbb{R}^3`

        .. math::

            \\text{cross}(\\boldsymbol{a},\\boldsymbol{b})=\\begin{pmatrix}
            a_2\\cdot b_3-a_3\\cdot b_2\\\\
            a_3\\cdot b_1-a_1\\cdot b_4\\\\
            a_1\\cdot b_2-a_2\\cdot b_1
            \\end{pmatrix}

        Parameters
        ----------
        vec_a : list
            First vector :math:`\\boldsymbol{a}`
        vec_b : list
            Second vector :math:`\\boldsymbol{b}`

        Returns
        -------
        vec : list
            Crossproduct
        """
        vec = []
        vec.append(vec_a[1]*vec_b[2]-vec_a[2]*vec_b[1])
        vec.append(vec_a[2]*vec_b[0]-vec_a[0]*vec_b[2])
        vec.append(vec_a[0]*vec_b[1]-vec_a[1]*vec_b[0])

        return vec

    def _angle_polar(self, pos, is_deg=False):
        """Calculate the polar angle of a position vector
        :math:`\\boldsymbol{a}\\in\\mathbb{R}^3`, which is the angle of the
        x-axis towards the reflected position vector on the x-y-plane

        .. math::

            \\text{polar}(\\boldsymbol{a})=\\arctan2(x,y)\\left\\{
            \\begin{array}{ll}
            \\tan^{-1}\\left(\\frac{y}{x}\\right)&x>0\\\\
            \\tan^{-1}\\left(\\frac{y}{x}\\right)+\\pi&x<0,y>0\\\\
            \\pm\\pi&x<0,y=0\\\\
            \\tan^{-1}\\left(\\frac{y}{x}\\right)-\\pi&x<0,y<0\\\\
            +\\frac{\\pi}{2}&x=0,y>0\\\\
            -\\frac{\\pi}{2}&x=0,y<0
            \\end{array}
            \\right.

        with :math:`x` as the first vector entry and :math:`y` as the second.

        Parameters
        ----------
        pos : list
            Position vector :math:`\\boldsymbol{a}`
        is_deg : bool, optional
            True if the output should be in degree

        Returns
        -------
        angle : float
            Polar angle
        """
        try:
            angle = math.atan2(pos[1], pos[0])
        except:
            angle = math.atan(0)

        return angle*180/math.pi if is_deg else angle

    def _angle_azi(self, pos, is_deg=False):
        """Calculate the azimuthal angle of a position vector
        :math:`\\boldsymbol{a}\\in\\mathbb{R}^3`, which is the angle of the
        position vector towards the x-y-plane

        .. math::

            \\text{azimut}(\\boldsymbol{a})=\\cos^{-1}\\frac{y}{|\\boldsymbol{a}|}

        with :math:`y` as the second vector entry.

        Parameters
        ----------
        pos : list
            Position vector :math:`\\boldsymbol{a}`
        is_deg : bool, optional
            True if the output should be in degree

        Returns
        -------
        angle : float
            Azimuthal angle
        """
        try:
            angle = math.acos(pos[2]/self._length(pos))
        except:
            angle = math.acos(0)

        return angle*180/math.pi if is_deg else angle

    def _axis(self, inp):
        """Return the three dimensional unit-vector of the main axes.
        Input is either integer or string

        * 1 or "x" for the x-axis
        * 2 or "y" for the y-axis
        * 3 or "z" for the z-axis

        Parameters
        ----------
        inp : integer, string
            Axis type input

        Returns
        -------
        vec : list
            Unit vector
        """
        # Error message
        axisError = "Wrong axis definition..."

        # Process input
        if isinstance(inp, str):
            if inp == "x":
                axis = 1
            elif inp == "y":
                axis = 2
            elif inp == "z":
                axis = 3
            else:
                return axisError
        elif isinstance(inp, int):
            if inp == 1 or inp == 2 or inp == 3:
                axis = inp
            else:
                return axisError
        else:
            return axisError

        # Return vector
        return [1 if i == axis-1 else 0 for i in range(self._dim)]

    def _focal(self):
        """Calculate the positional focal-point

        .. math::

            \\text{focal}=\\begin{pmatrix}f_1&f_2&\\dots&f_n\\end{pmatrix}^T

        with

        .. math::

            f_i=\\frac{1}{m}\\sum_j^m d_{ij}

        Returns
        -------
        focal : list
            Focal point
        """
        return [sum(self._data[i])/len(self._data[i]) for i in range(self._dim)]

    def _rotate(self, data, axis, angle, is_deg):
        """Rotate a vector :math:`\\boldsymbol{a}\\in\\mathbb{R}^3`
        along an axis :math:`\\boldsymbol{b}\\in\\mathbb{R}^3` with angle
        :math:`\\alpha\\in\\mathbb{R}`.
        Input for the axis is either a vector or the input for
        function :func:`_vector`.
        The rotation is performed using the rotation-matrix

        .. math::

            \\boldsymbol{R}_\\boldsymbol{n}(\\alpha)=\\begin{pmatrix}
            n_1^2 (1-\\cos\\alpha)+   \\cos\\alpha&n_1n_2(1-\\cos\\alpha)-n_3\\sin\\alpha&n_1n_3(1-\\cos\\alpha)+n_2\\sin\\alpha\\\\
            n_2n_1(1-\\cos\\alpha)+n_3\\sin\\alpha&n_2^2 (1-\\cos\\alpha)+   \\cos\\alpha&n_2n_3(1-\\cos\\alpha)-n_1\\sin\\alpha\\\\
            n_3n_1(1-\\cos\\alpha)-n_2\\sin\\alpha&n_3n_2(1-\\cos\\alpha)+n_1\\sin\\alpha&n_3^2 (1-\\cos\\alpha)+   \\cos\\alpha
            \\end{pmatrix}


        where :math:`n_i` are the entries for the unit vector
        :math:`\\boldsymbol{n}` of the axis. The new coordinates :math:`\\boldsymbol{c}`
        are then calculated using a matrix-vector multiplication

        .. math::

            \\boldsymbol{c}=\\boldsymbol{R}_\\boldsymbol{n}\\boldsymbol{a}.

        Parameters
        ----------
        data : list
            Vector :math:`\\boldsymbol{a}`
        axis : integer, string, list
            Axis :math:`\\boldsymbol{b}`
        angle : float
            Angle
        is_deg : bool
            True if the input is in degree

        Returns
        -------
        coord : list
            Vector c as the result of the rotation
        """
        # Angle
        angle = angle*math.pi/180 if is_deg else angle

        # Set vector
        if isinstance(axis, list):
            if len(axis) == self._dim:
                n = axis
            elif len(axis) == 2:
                n = self._vector(axis[0], axis[1])
            else:
                print("Wrong vector dimensions.")
                return
        else:
            n = self._axis(axis)
            if isinstance(n, str):
                print(n)
                return

        # Unit vector
        n = self._unit(n)

        # Define rotation matrix
        n1 = n[0]
        n2 = n[1]
        n3 = n[2]

        c = math.cos(angle)
        s = math.sin(angle)

        r = [[n1*n1*(1.-c) + c, n1*n2*(1.-c) - n3*s,  n1*n3*(1.-c) + n2*s],
             [n2*n1*(1.-c) + n3*s, n2*n2*(1.-c) + c,  n2*n3*(1.-c) - n1*s],
             [n3*n1*(1.-c) - n2*s, n3*n2*(1.-c) + n1*s,  n3*n3*(1.-c) + c]]

        # Rotate
        return [data[0]*r[i][0]+data[1]*r[i][1]+data[2]*r[i][2] for i in range(self._dim)]

    def _box_size(self):
        """Calculate the boxsize of the current molecule. This is done by
        determining the maximal coordinate value of all atoms in all dimensions

        .. math::

            \\boldsymbol{b}=\\begin{pmatrix}\\max(\\boldsymbol{d}_1)&max(\\boldsymbol{d}_1)&\\dots&max(\\boldsymbol{d}_n)\\end{pmatrix}^T

        where :math:`\\boldsymbol{d}_i` is the dimension-vector of the data matrix.

        Returns
        -------
        box : list
            Box length of the current molecule
        """
        return [max(self._data[i]) if max(self._data[i]) > 0 else 0.001 for i in range(self._dim)]


    ##############################
    # Public Methods - Transform #
    ##############################
    def translate(self, vec):
        """Translate data matrix :math:`\\boldsymbol{D}` along a vector
        :math:`\\boldsymbol{a}\\in\\mathbb{R}^n`.

        .. math::

            \\boldsymbol{D}_\\text{trans}=
            \\boldsymbol{D}+\\boldsymbol{a}=
            \\begin{pmatrix}
            \\boldsymbol{d}_1+a_1&\\boldsymbol{d}_2+a_2&\\dots&\\boldsymbol{d}_n+a_n&\\boldsymbol{d}_t
            \\end{pmatrix}

        Parameters
        ----------
        vec : list
            Vector a
        """
        # Translate coordinates
        for i in range(self._dim):
            for j in range(len(self._data[i])):
                self._data[i][j] += vec[i]

    def rotate(self, axis, angle, is_deg=True):
        """Rotate data matrix :math:`\\boldsymbol{D}` around an axis
        :math:`\\boldsymbol{a}\\in\\mathbb{R}^3` with angle
        :math:`\\alpha\\in\\mathbb{R}` using rotation function :func:`_rotate`.

        Parameters
        ----------
        axis : integer, string, list
            Axis
        angle : float
            Angle
        is_deg : bool, optional
            True if the input is in degree
        """
        # Initialize
        dim = self._dim
        data = self._data

        # Rotate
        for i in range(len(data[0])):
            coord = [0 for i in range(dim)]
            for j in range(dim):
                coord[j] += data[j][i]

            coord = self._rotate(coord, axis, angle, is_deg)
            for j in range(dim):
                data[j][i] = coord[j]

    def put(self, atom, pos):
        """Change the position of an atom to a given vector
        :math:`\\boldsymbol{a}\\in\\mathbb{R}^n`.

        Parameters
        ----------
        atom : integer, string
            Atom-id whose position will be changed, use **"last"** for the newest atom
        pos : list
            New position vector
        """
        # User input
        atom = len(self._data[0])-1 if atom == "last" else atom

        # Change position
        for i in range(len(pos)):
            self._data[i][atom] = pos[i]

    def move(self, atom, pos):
        """Move whole the molecule to a new position, where the dragging point
        is a given atom that is moved to a specified position.

        Parameters
        ----------
        atom : integer, string
            Main atom-id whose position will be changed, use **"last"** for the newest atom
        pos : list
            New position vector
        """
        # Calculate vector
        if atom == "last":
            atom = len(self._data[0])-1

        vec = self._vector(self.pos(atom), pos)

        # Translate
        self.translate(vec)

    def zero(self, pos=[0, 0, 0]):
        """Move whole the molecule, so that the minimal coordinate
        between all atoms is zero in all dimensions, or rather the values of the
        position variable ``pos``. This function is basically setting
        the zero point of the coordinate system to ``pos``.

        Parameters
        ----------
        pos : list, optional
            Vector of the zero point of the coordinate system

        Returns
        -------
        vec : list
            Vector used for the translation
        """
        # Calculate translation vector
        vec = [pos[i]-min(self._data[i]) for i in range(self._dim)]

        # Resett box size
        self._box = None

        # Translate molecule
        self.translate(vec)

        return vec

    def box(self, size, is_extend=True):
        """Create a molecule box around a centered molecule, where the
        coordinate systems zero point is set to the positional focal point of
        the molecule. The resulting box has the edge length of the given size
        value. If ``is_extend`` is set to True, then the maximal molecule
        coordinate of all dimensions is added to ``size``.

        Parameters
        ----------
        size : float
            Box size
        is_extend : bool, optional
            Should the molecule size be added to the box-size
        """
        # Get focal point
        focal = self._focal()

        # Calculate box size
        boxFocal = [f+size for f in focal]
        self._box = [max(boxFocal) if is_extend else size for f in range(3)]

        # Move molecule to box center
        self.add("R", focal)
        self.move("last", [b/2 for b in self._box])
        self.delete("last")


    #########################
    # Public Methods - Edit #
    #########################
    def part_move(self, bond, atoms, length, vec=None):
        """Change the length of a specified bond. Variable ``atoms`` specifies
        which atoms or rather which part of the molecule needs to be moved for
        this specific bond. The given length is going to be the new bond length,
        not by how much the bond length is changed.

        The move vector is determined automatically by the given length and atom bond.
        This vector can also be given manually with no regards to length,
        by setting the vector ``vec``.

        Parameters
        ----------
        bond : list
            List of two atom ids of the bond to be adjusted
        atoms : integer, list
            List of atoms that need to be moved by changing the bond length (can also be one id)
        length : float
            New bond length
        vec : list, optional
            Set this vector, to manually move the atoms

        Examples
        --------
        >>> part_move([0,1],[1,2,3],0.5)
        """
        # Create temporary molecule
        if isinstance(atoms, int):
            atoms = [atoms]
        temp = self._temp(atoms)

        # Set length
        length = abs(length-self.bond(*bond)[1])

        # Set vector
        if vec == None:
            vec = self._vector(bond[0], bond[1])
        vec = [v*length for v in self._unit(vec)]

        # Move molecule
        temp.translate(vec)

        # Update positions
        self._update(temp.get_data(), atoms)

    def part_rotate(self, bond, atoms, angle, zero):
        """Rotate a set of specified atoms around a given bond as the rotation
        axis. First however the system needs to be set to zero. Therefore the
        atom id to define the new coordinate system has to be given for the set
        of specified atoms. Normally this is the atoms that lies at on an end
        of the given bond axis.

        Parameters
        ----------

        bond : list
            List of two atom ids of the bond to be set as an axis
        atoms : integer, list
            List of atoms to be rotated (can also be one id)
        angle : float
            Rotation angle
        zero : integer
            Atom id to define zero point of the new coordinate system

        Examples
        --------
        >>> part_rotate([0,1],[1,2,3],90,0)
        """
        # Create temporary molecule
        self.move(zero, [0, 0, 0])
        if isinstance(atoms, int):
            atoms = [atoms]
        temp = self._temp(atoms)

        # Rotate molecule
        temp.rotate([self.pos(bond[0]), self.pos(bond[1])], angle)

        # Update positions
        self._update(temp.get_data(), atoms)

    def part_angle(self, bond_a, bond_b, atoms, angle, zero):
        """Change the bond angle of two bond vectors. Variable ``atoms`` specifies
        which atoms or rather which part of the molecule needs rotated to change
        the specified bond angle. First however the system needs to be set to
        zero. Therefore the atom id to define the new coordinate system has to be
        given for the set of specified atoms. Normally this is the atom that
        touches the angle.

        The rotation axis is determined by creating the cross product
        of the two bond vectors. Thus getting the normal vector of a surface
        that contains both bond vectors.

        Parameters
        ----------
        bond_a : list
            First bond vector given as a list of two atom ids
        bond_b : list
            Second bond vector given as a list of two atom ids
        atoms : integer, list
            List of atoms to be rotated (can also be one id)
        angle : float
            Rotation angle
        zero : integer
            Atom id to define zero point of the new coordinate system

        Examples
        --------
        >>> part_angle([0,1],[1,2],[1,2,3],90,1)
        """
        # Create temporary molecule
        self.move(zero, [0, 0, 0])
        if isinstance(atoms, int):
            atoms = [atoms]
        temp = self._temp(atoms)

        # Rotate molecule around normal vector
        if len(bond_a) == 2 and len(bond_b) == 2:
            vec = self._cross(self._vector(bond_a[0], bond_a[1]),
                              self._vector(bond_b[0], bond_b[1]))
        elif len(bond_a) == self._dim and len(bond_b) == self._dim:
            vec = self._cross(bond_a, bond_b)
        else:
            print("Wrong bond input...")
            return
        temp.rotate(vec, angle)

        # Update positions
        self._update(temp.get_data(), atoms)

    def length_angle(self, bond, bond_a, bond_b, atoms, zero, length, angle=[0, 45], grid=0.001, is_silent=True, is_negative=False):
        """If the bond length of an exemplary circular molecule is correlated
        to an angle, then the bond length has to be changed in dependence to
        the bond angle.

        Variable ``atoms`` specifies which atoms or rather which part of the
        molecule needs moved and rotated to change the specified bond length.

        First however the system needs to be set to zero. Therefore the atom id
        to define the new coordinate system has to be given for the set of
        specified atoms. Normally this is the atom that touches the angle.

        The rotation axis is determined by creating the cross product
        of the two bond vectors. Thus getting the normal vector of a surface
        that contains both bond vectors.

        The new length is determined, by rotating the specified molecule part in
        steps and calculating the resulting length error. If the error is zero,
        then the rotation steps stop.

        Parameters
        ----------
        bond : list
            Bond of which the length should be changed
        bond_a : list
            First bond vector given as a list of two atom ids
        bond_b : list
            Second bond vector given as a list of two atom ids
        atoms : integer, list
            List of atoms to be rotated (can also be one id)
        zero : integer
            Atom id to define zero point of the new coordinate system
        length : float
            New bond length
        angle : list, optional
            Angle range to be tested
        grid : float, optional
            angle steps fineness
        is_silent : bool, optional
            True to suppress error-value messages
        is_negative : bool, optional
            True for a negative rotation

        Examples
        --------
        >>> length_angle([0,19],[8,9],[9,10],[10,11,12],self.pos(9))

        """
        angles = [x*grid for x in range(int(angle[0]/grid), int(angle[1]/grid)+1)]
        for a in angles:
            a = -a if is_negative else a
            self.part_angle(bond_a, bond_b, atoms, a, zero)

            error = round(abs(self.bond(*bond)[1]-length), 3)

            if not is_silent:
                print("Error = "+str(error))

            if error == 0:
                break


    #############################
    # Public Methods - Calulate #
    #############################
    def angle(self, vec_a, vec_b, is_deg=True):
        """Calculate the angle between two vectors
        :math:`\\boldsymbol{a},\\boldsymbol{b}\\in\\mathbb{R}^n`

        .. math::

            \\text{angle}=\\cos^{-1}\\frac{\\boldsymbol{a}\cdot\\boldsymbol{b}}
            {|\\boldsymbol{a}||\\boldsymbol{a}|}

        Parameters
        ----------
        vec_a : list
            First vector :math:`\\boldsymbol{a}`
        vec_b : list
            Second vector :math:`\\boldsymbol{b}`
        is_deg : bool, optional
            True if the output should be in degree

        Returns
        -------
        angle : float
            Angle
        """
        angle = math.acos(self._dotproduct(vec_a, vec_b)/(self._length(vec_a)*self._length(vec_b)))

        return angle*180/math.pi if is_deg else angle

    def pos(self, atom):
        """Get the position of an atom as a vector.

        Parameters
        ----------
        atom : integer
            Atom id

        Returns
        -------
        pos : list
            Position vector of the specified atom
        """
        return [self._data[i][atom] for i in range(self._dim)]

    def bond(self, inp_a, inp_b):
        """Return the bond length and vector of a specified bond. The two inputs
        can either be atom indices or to vectorial positions.

        Parameters
        ----------
        inp_a : integer, list
            Either an atom id or a position vector
        inp_b : integer, list
            Either an atom id or a position vector

        Returns
        -------
        bond : list
            Bond vector and length

        Examples
        --------
        >>> bond(0,1)
        >>> bond(*[0,1])
        >>> bond([1,0,0],[0,0,0])
        """
        return [self._vector(inp_a, inp_b), self._length(self._vector(inp_a, inp_b))]


    #########################
    # Public Methods - Edit #
    #########################
    def add(self, name, pos, bond=None, r=0, theta=0, phi=0, is_deg=True):
        """Add a new atom in polar coordinates. The ``pos`` input is either
        an atom id, that determines is the bond-start,
        or a vector for a specific position.

        Bond has to be given, if the polar coordinates are dependet on the bond
        vector as the basic axis. The coordinate system is then transformed
        to the new bond axis. If set to None, then the given coordinates are
        thought to be dependent on the basic axes.

        If the given position is an atom id, then a bond is automatically added
        to the bond matrix.

        Parameters
        ----------
        name : string
            Atom type
        pos : integer, list
            Position of the atom
        bond : None, list, optional
            Bond axis
        r : float, optional
            Bond length
        theta : float, optional
            Azimuthal angle
        phi : float, optional
            Polar angle
        is_deg : bool, optional
            True if the input of the angles in degree

        Examples
        --------
        >>> add("C",[0,0,0])
        >>> add("C",0,r=0.153,theta=-135)
        >>> add("C",1,[0,1],r=0.153,theta= 135)
        """
        # Angles
        phi *= math.pi/180 if is_deg else 1
        theta *= math.pi/180 if is_deg else 1

        # Transform sphercial to cartesian coordinates
        x = r*math.sin(theta)*math.cos(phi)
        y = r*math.sin(theta)*math.sin(phi)
        z = r*math.cos(theta)
        coord = [x, y, z]

        # Bond vector
        if bond == None:
            vec = self._axis("z")
        else:
            vec = self._vector(bond[0], bond[1])

        # Calculate angles for rotation
        phi = self._angle_polar(vec)
        theta = self._angle_azi(vec)

        # Rotate towards axis
        tRot = self._cross([0, 0, 1], vec)
        if sum(tRot) == 0:
            tRot = "y"

        coord = self._rotate(coord, tRot, theta, is_deg=False)

        # Process position input
        b = None
        if isinstance(pos, int):
            b = pos
            pos = self.pos(pos)

        # Add coordinates
        for i in range(self._dim):
            self._data[i].append(pos[i]+coord[i])

        # Add name
        self._data[self._dim].append(name)

    # Delete an atom
    def delete(self, atom):
        """Delete specified atom from the molecule.
        The input can also be a list of atom ids.

        Parameters
        ----------
        atom :  int, list
            Atom id or list to be deleted
        """
        # Process input
        if isinstance(atom, str):
            if atom == "last":
                atom = [len(self._data[0])-1]
        elif not isinstance(atom, list):
            atom = [atom]
        atom = sorted(atom, reverse=True)

        # Remove line
        for data in self._data:
            for a in atom:
                data.pop(a)

    def overlap(self, error=0.005, is_print=False):
        """Check if atoms are overlapping (as in duplicates) and delete these.

        Parameters
        ----------
        error : float, optional
            Error of the overlap search
        is_print : bool, optional
            True to print the list of overlapping atoms
        """
        # Initialize
        data = self._data
        dim = self._dim

        # Find duplicate atoms in reverse
        for i in list(reversed(range(self.get_num()))):
            # Get positions
            atom_a = self.pos(i)
            for j in range(self.get_num()):
                if not j == i:
                    atom_b = self.pos(j)

                    # Check if identical
                    sum_del = 0
                    a = []
                    for k in range(dim):
                        if abs(atom_a[k]-atom_b[k]) < error:
                            a.append(abs(atom_a[k]-atom_b[k]))
                            sum_del += 1

                    # Delete atom
                    if sum_del == dim:
                        # print(i,j,a)
                        self.delete(i)
                        if is_print:
                            print(i)
                        break

    def change_type(self, atom, atom_type):
        """Change the atom type of a specified atom.

        Parameters
        ----------
        atom : integer
            Atom id
        atom_type : string
            New atomtype
        """
        self._data[self._dim][atom] = atom_type

    def switch_order(self, atom_a, atom_b):
        """Switch atom order of two atoms.

        Parameters
        ----------
        atom_a : integer
            Atom id of the first atom
        atom_b : integer
            Atom id of the second atom
        """
        # Get data
        data_a = [self._data[i][atom_a] for i in range(self._dim+1)]
        data_b = [self._data[i][atom_b] for i in range(self._dim+1)]

        # Switch data
        for i in range(self._dim+1):
            self._data[i][atom_a] = data_b[i]
            self._data[i][atom_b] = data_a[i]


    ###########################
    # Public Methods - Output #
    ###########################
    def table(self):
        """Create a pandas table of the molecule data."""
        import pandas as pd

        # Create dictionary
        data = {"Axis "+str(dim+1): self._data[dim] for dim in range(self._dim)}
        data["Type"] = self._data[-1]

        # Create dataframe
        return pd.DataFrame(data)


    ##################
    # Setter Methods #
    ##################
    def set_name(self, name):
        """Set the molecule name.

        Parameters
        ----------
        name : string
            Molecule name
        """
        self._name = name

    def set_short(self, short):
        """Set the molecule short name

        Parameters
        ----------
        short : string
            Molecule short name
        """
        self._short = short

    def set_mol_list(self, mol_list):
        """Set the molecule list for writing the structure file

        Parameters
        ----------
        mol_list : list
            List of molecule objects
        """
        self._mol_list = mol_list

    def set_box(self, box):
        """Set the molecule box dimensions.

        Parameters
        ----------
        box : list
            Box dimension
        """
        self._box = box

    def set_charge(self, charge):
        """Set the total charge of the molecule.

        Parameters
        ----------
        charge : float
            Total molecule charge
        """
        self._charge = charge

    def set_masses(self, masses=None):
        """Set the molar masses of the atoms.

        Parameters
        ----------
        masses : list, None, optional
            List of molar masses in :math:`\\frac g{mol}`
        """
        import porems.database as db
        self._masses = [db.get_mass(atom) for atom in self._data[self._dim]] if masses is None else masses

    def set_mass(self, mass=None):
        """Set the molar mass of the molecule.

        Parameters
        ----------
        mass : float, None, optional
            Molar mass in :math:`\\frac g{mol}`
        """
        self._mass = sum(self.get_masses()) if mass is None else mass

    def set_com(self, com=None):
        """Set the center of mass coordinates of the molecule.

        Parameters
        ----------
        com : list, None, optional
            Center of mass coordinates
        """
        if com is not None:
            self._com = com
        else:
            self._com = []
            masses = self.get_masses()
            for i in range(self._dim):
                self._com.append(0)
                for j in range(len(self._data[i])):
                    self._com[i] += masses[j]*self._data[i][j]

            self._com = [x/sum(masses) for x in self._com]


    ##################
    # Getter Methods #
    ##################
    def get_data(self):
        """Return the data matrix.

        Returns
        -------
        data : list
            Data matrix
        """
        return self._data

    def get_name(self):
        """Return the molecule name.

        Returns
        -------
        name : string
            Molecule name
        """
        return self._name

    def get_short(self):
        """Return the molecule short name.

        Returns
        -------
        short : string
            Molecule short name
        """
        return self._short

    def get_mol_list(self):
        """Return the global molecule list for writing the structure file.

        Returns
        -------
        mol_list : list
            List of all molecules
        """
        return self._mol_list

    def get_box(self):
        """Calculate and return the box size of the current molecule.

        Returns
        -------
        box : list
            Box dimensions
        """
        return self._box_size()

    def get_box_c(self):
        """Return the current box size.

        Returns
        -------
        box : list
            Box dimensions
        """
        return self._box

    def get_num(self):
        """Return the number of atoms.

        Returns
        -------
        num : integer
            Number of atoms
        """
        return len(self._data[0])

    def get_type(self, atom):
        """Return the atomtype of the given atom id.

        Parameters
        ----------
        atom : integer
            Atom id

        Returns
        -------
        box : list
            Box dimensions
        """
        return self._data[self._dim][atom]

    def get_charge(self):
        """Return the total charge of the molecule.

        Returns
        -------
        charge : float
            Total charge
        """
        return self._charge

    def get_masses(self):
        """Return a list of masses of the atoms.

        Returns
        -------
        masses : list
            Masses in :math:`\\frac g{mol}`
        """
        if self._masses is None:
            self.set_masses()
        return self._masses

    def get_mass(self):
        """Return the molar mass of the molecule.

        Returns
        -------
        mass : float
            Molar mass in :math:`\\frac g{mol}`
        """
        if self._mass is None:
            self.set_mass()
        return self._mass

    def get_com(self):
        """Return the center of mass of the molecule.

        Returns
        -------
        com : list
            center of mass coordinates
        """
        if self._com is None:
            self.set_com()
        return self._com



################################################################################
# Verlet Class                                                                 #
#                                                                              #
"""Verlet lists definitions."""
################################################################################


import math
import multiprocessing as mp

import porems.utils as utils


class Verlet:
    """This class converts the molecule box into verlet lists.

    The aim of verlet lists is reducing the workload on search algorithms for
    atom pairs. Normally the effort would be

    .. math::

        \\mathcal O(n^2)

    with the number of atoms :math:`n`, since each atom has to be compared with
    each atom if their bond length is within the search distance.
    Since this distance is however confined, the idea is reducing
    the search space by dividing the molecule box into small verlet boxes. Thus the
    atom pairs have to only be checked in the box containing the starting atom
    and the small boxes surrounding the main box. In the three dimensional case,
    the effort reduces to

    .. math::

        \\begin{array}{cc}
            \\mathcal O(n\\cdot N),&N=\\sum_{j=0}^ma_j
        \\end{array}

    with the number of boxes m, in the three dimensional case this is :math:`3^3=27`,
    the number of atoms :math:`a_j` of box :math:`j` and the number of surround
    atoms :math:`N`. This is considerably less and of course depending on the
    verlet box size. Note that the verlet size :math:`v` has to be greater than the
    specified bond length :math:`l` for the search

    .. math::

        v > l.

    Since molecular simulation considers periodic
    boundary conditions, these have to also be applied for the surround box search.

    Parameters
    ----------
    mol : Molecule
        Molecule to be divided
    size : float
        Verlet box edge size
    is_pbc : bool
        True if periodic boundry conditions are needed
    """
    def __init__(self, mol, size, is_pbc):
        # Initialize
        self._dim = 3
        self._np = mp.cpu_count()
        self._is_pbc = is_pbc
        self._mol = mol
        self._data = mol.get_data()
        self._box = [[], []]

        self._verlet(size)
        self._fill()


    ################################
    # Private Methods - Definition #
    ################################
    def _verlet(self, size):
        """Here the possible number of verlet boxes is calculated for each dimension
        for the given size and molecule dimensions. The given size is adjusted,
        since only whole numbered box number are reasonable.
        A list of boxes is generated containing the coordinates of the zero
        point of each box. Furthermore an empty list for each box added, that will
        contain pointer to the atoms.

        The boxes are numbered as follows for the example of a :math:`3\\times3` block

        .. math::

            \\begin{array}{ccc}
                \\begin{bmatrix}
                     6& 7& 8\\\\
                     3& 4& 5\\\\
                     0& 1& 2
                \\end{bmatrix}\\text{ for }z=0,&
                \\begin{bmatrix}
                    15&16&17\\\\
                    12&13&14\\\\
                     9&10&11
                \\end{bmatrix}\\text{ for }z=1,\dots
            \\end{array}

        where the x-axis is the first count direction, y-axis the second and
        z-axis the third. The position list for a box id contains the values for
        the x, y and z values. For example the position of box id :math:`12`
        would be

        .. math::

            \\text{pos}(12)=\\begin{bmatrix}0&1&1\\end{bmatrix}.

        Parameters
        ----------
        size : float
            Desired verlet box length
        """
        # Initialize
        boxes = []
        box_size = self._mol.get_box()

        # Calculate box number and sizes
        count = [math.floor(box/size) for box in box_size]
        size = [box_size[i]/count[i] for i in range(len(count))]

        # Fill verlet list - k=x, j=y, i=z
        for i in range(count[2]):
            for j in range(count[1]):
                for k in range(count[0]):
                    boxes.append([size[0]*k, size[1]*j, size[2]*i])

        # Globalize
        self._box[0] = boxes
        self._size = size
        self._count = count

    def _fill(self):
        """Atoms are so to say filled in the verlet boxes depending on their position.
        This is done by adding atom ids to specific boxes in the box array.
        """
        # Initialize
        data = self._data
        data = self._mol.get_data()
        dim = self._dim
        size = self._size
        count = self._count
        pointer = [[] for i in range(count[0]*count[1]*count[2])]

        # Fill boxes with atoms
        for i in range(len(data[0])):
            pos = [0 for i in range(dim)]
            for j in range(dim):
                pos[j] = math.floor(data[j][i]/size[j])
                pos[j] -= 1 if pos[j] == count[j] else 0

            pointer[self._index(pos)].append(i)

        # Globalize
        self._box[1] = pointer


    ##############################
    # Private Methods - Iterator #
    ##############################
    def _index(self, pos):
        """Get the box id from a given box position which is a list of x, y and z values.
        
        Parameters
        ----------
        pos : list
            Box position in x, y and z values

        Returns
        -------
        index : integer
            Box index
        """
        # Initialize
        count = self._count
        index = 0

        # Calculate id
        index += pos[0]
        index += pos[1]*count[0]
        index += pos[2]*count[0]*count[1]

        return int(index)

    def _position(self, index):
        """Get the box position in x, y and z values from a given box id.

        Parameters
        ----------
        index : integer
            Box index

        Returns
        -------
        pos : list
            Box position in x, y and z values
        """
        # Initialize
        dim = self._dim
        count = self._count
        pos = [0 for i in range(self._dim)]

        # Get index
        pos[2] = math.floor(index/(count[0]*count[1]))
        pos[1] = math.floor((index-pos[2]*count[0]*count[1])/count[0])
        pos[0] = index-pos[1]*count[0]-pos[2]*count[0]*count[1]

        return pos

    def _input(self, inp):
        """Check if input is a box id or a position. If it is an id then
        the id is transformed to a position list.

        Parameters
        ----------
        inp : integer, list
            Input

        Returns
        -------
        pos : list
            Box position in x,y and z values
        is_id : bool
            True if the input was an id
        """
        if isinstance(inp, list):
            pos = inp
            is_id = False
        elif isinstance(inp, int):
            pos = self._position(inp)
            is_id = True
        else:
            print("Wrong input!")
            return None

        return pos, is_id

    def _forward(self, dim, inp):
        """Helper function for stepping forward with the iterator. Periodic
        boundary conditions are applied depending on the global pbc variable.

        Parameters
        ----------
        dim : integer
            Stepping dimension
        inp : integer, list
            Position input, can also be an id

        Returns
        -------
        pos : integer, list
            Iterated position in the input format
        """
        # Check if input is None
        if inp is None:
            return

        # Process input
        pos, is_id = self._input(inp)

        # Move
        if pos[dim] == self._count[dim]-1:
            pos[dim] = 0 if self._is_pbc else None
        else:
            pos[dim] += 1

        # Return
        if not None in pos:
            return self._index(pos) if is_id else pos

    def _backward(self, dim, inp):
        """Helper function for stepping backward with the iterator. Periodic
        boundary conditions are applied depending on the global pbc variable.

        Parameters
        ----------
        dim : integer
            Stepping dimension
        inp : integer, list
            Position input, can also be an id

        Returns
        -------
        pos : integer, list
            Iterated position in the input format
        """
        # Check if input is None
        if inp is None:
            return

        # Process input
        pos, is_id = self._input(inp)

        # Move
        if pos[dim] == 0:
            pos[dim] = self._count[dim]-1 if self._is_pbc else None
        else:
            pos[dim] -= 1

        # Return
        if not None in pos:
            return self._index(pos) if is_id else pos

    # Iterators
    def _right(self, inp): return self._forward(0, inp)
    def _left(self, inp): return self._backward(0, inp)
    def _top(self, inp): return self._forward(1, inp)
    def _bot(self, inp): return self._backward(1, inp)
    def _back(self, inp): return self._forward(2, inp)
    def _front(self, inp): return self._backward(2, inp)


    ###########################
    # Parallelization methods #
    ###########################
    def _find_bond(self, box_list, atom_type, distance, error, condition):
        """Search for a bond in the given verlet boxes. This function automatically
        searches for the surrounding boxes and checks in all 27 of them for the
        defined bond (atomtypes and bond length). If the bond length is within a
        given error, then the atoms will be added to an output.
        The user can also define an additional and stricter condition for adding
        the atoms.

        Parameters
        ----------
        box_list : None, list
            None for all verlet boxes, list for specified boxes
        atom_type : list
            List of two atom type stings
        distance : float
            Bond length
        error : float
            Tollerated deviation of the bondlength
        condition : None, func
            None or a function f(Molecule,bond) with a boolean output

        Returns
        -------
        bonds : list
            Bond array containing lists of two atom ids
        """
        # Initialize
        data = self._data
        dim = self._dim
        size = self._size
        count = self._count
        box_len = [count[i]*size[i] for i in range(dim)]
        output = []

        # User input
        if box_list == None:
            box_list = [i for i in range(len(self.get_box()[1]))]

        # Loop through all boxes
        for box in box_list:
            # Get surrounding boxes
            neighbour = self.neighbour(box)
            atoms = []
            for n in neighbour:
                atoms.extend(self._box[1][n])

            # Run through atoms in main boxes
            for atom_a in self._box[1][box]:
                # Check type
                if data[3][atom_a] == atom_type[0]:
                    entry = [atom_a, []]
                    pos_a = self._mol.pos(atom_a)
                    # Search in neighbouring boxes for partner
                    for atom_b in atoms:
                        if data[3][atom_b] == atom_type[1] and not atom_a == atom_b:
                            # Calculate length
                            length = self._mol.bond(atom_a, atom_b)[1]

                            # Periodic boundaries
                            if length > max(size)*3 and self._is_pbc:
                                pos_b = self._mol.pos(atom_b)
                                temp = [pos for pos in pos_b]

                                for i in range(dim):
                                    # Without z-axis
                                    if abs(pos_b[i]-pos_a[i]) > size[i]*3:  # and not i==dim-1:
                                        temp[i] += -box_len[i] if pos_b[i] > pos_a[i] else box_len[i]

                                # New length
                                length = self._mol.bond(self._mol.pos(atom_a), temp)[1]

                            # Add if in range
                            if abs(length-distance) < error:
                                entry[1].append(atom_b)

                    # Add atom
                    is_add = True
                    if condition is not None:
                        is_add = condition(self._mol, entry[0])
                    if is_add:
                        output.append(entry)

        return output

    def find_parallel(self, box_list, atom_type, distance, error, condition=None, is_time=False):
        """Parallelized bond search of function :func:`_find_bond`.

        Parameters
        ----------
        box_list : None, list
            None for all verlet boxes, list for specified boxes
        atom_type : list
            List of two atom type stings
        distance : float
            Bond length
        error : float
            Tollerated deviation of the bondlength
        condition : None, function, optional
            None or a function f(Molecule,bond) with a boolean output
        is_time : bool, optional
            True to print the used search time

        Returns
        -------
        bonds : list
            Bond array containing lists of two atom ids
        """
        # Initialize
        np = self._np

        # User input
        if box_list == None:
            box_list = [i for i in range(len(self.get_box()[1]))]

        # Time calculation
        if is_time:
            t = m.tic()

        # Break box list down on number of processors
        box_len = math.floor(len(box_list)/np)

        boxes = []
        for i in range(np):
            if i == np-1:
                boxes.append(box_list[box_len*i:])
            else:
                boxes.append(box_list[box_len*i:box_len*(i+1)])

        # Parallel search
        pool = mp.Pool(processes=np)
        results = [pool.apply_async(self._find_bond, args=(
            x, atom_type, distance, error, condition)) for x in boxes]

        # Concat processes
        output = []
        for p in results:
            output.extend(p.get())

        # Time calculation
        if is_time:
            m.toc(t, "Find bond "+atom_type[0]+"-"+atom_type[1])

        return output


    ##################
    # Public methods #
    ##################
    def neighbour(self, index, is_self=True):
        """Get the ids of the boxes surrounding the given one.

        Parameters
        ----------
        index : integer
            Main box id
        is_self : bool, optional
            True to add the main box to the output

        Returns
        -------
        neighbour : list
            List of surrounding boxes
        """
        # Initialize
        neighbour = []

        # Find neighbours
        z = [self._back(index), index, self._front(index)]
        y = [[self._top(i), i, self._bot(i)] for i in z]

        for i in range(len(z)):
            for j in range(len(y[i])):
                neighbour.append(self._left(y[i][j]))
                neighbour.append(y[i][j])
                neighbour.append(self._right(y[i][j]))

        if not is_self:
            neighbour.pop(13)

        return [n for n in neighbour if n is not None]

    def delete(self, inp):
        """Delete given atoms, then refill the verlet boxes,
        since the list ids changed.

        Parameters
        ----------
        inp : list, integer
            Atom list or single atom id
        """
        self._mol.delete(inp)
        self._fill()


    ##################
    # Setter methods #
    ##################
    def set_pbc(self, pbc):
        """Turn the periodic boundary consitions on and off.

        Parameters
        ----------
        bond : bool
            PBC mode
        """
        self._is_pbc = pbc


    ##################
    # Getter methods #
    ##################

    def get_box(self):
        """Return the box list.

        Returns
        -------
        box : list
            Box list
        """
        return self._box

    def get_count(self):
        """Return the number of boxes in each dimension.

        Returns
        -------
        count : list
            Box number
        """
        return self._count

    def get_size(self):
        """Return the box size in each dimension.

        Returns
        -------
        size : list
            Box size
        """
        return self._size

    def get_mol(self):
        """Return the molecule.

        Returns
        -------
        mol : Molecule
            Molecule
        """
        return self._mol



################################################################################
# Bonding Class                                                                #
#                                                                              #
"""Bonding search structure."""
################################################################################


import math

import porems.utils as utils


class Bonding:
    """The aim of this class is preserving all information of the silicon grid
    bonds. This is needed since firstly, verlet searches get costly the more often
    they are needed and secondly, multiple searches deteriorate the bond information,
    due to numeric errors thus resulting in bonds beeing lost.

    The idea here was reducing the verlet searches to a single search by creating
    a connection matrix of all oxygene and silicon atoms.
    In fact two matrices :math:`ox\\in\\mathbb{N}^{no\\times2}` for oxygene and
    :math:`si\\in\\mathbb{N}^{ns\\times2}` were created with numbers of
    oxygene atoms :math:`no` and silicon atoms :math:`ns`

    .. math::

        \\begin{array}{cc}
            ox = \\left[
                \\begin{array}{ll}
                    o_0&[s_{0,0},s_{0,1}]\\\\
                    o_1&[s_{1,0},s_{1,1}]\\\\
                    &\\vdots\\\\
                    o_{no}&[s_{no,0},s_{no,1}]
                \\end{array}
            \\right],&
            si = \\left[
                \\begin{array}{ll}
                    s_0&[o_{0,0},o_{0,1},o_{0,2},o_{0,3}]\\\\
                    s_1&[o_{1,0},o_{1,1},o_{1,2},o_{1,3}]\\\\
                    &\\vdots\\\\
                    s_{ns}&[o_{ns,0},o_{ns,1},o_{ns,2},o_{ns,3}]
                \\end{array}
            \\right]
        \\end{array}

    with oxygene :math:`o_{i=0,\\dots,no}` and silicon :math:`s_{j=0,\\dots,ns}`.
    The entries next to the main atoms are the silicon atoms :math:`s_{i,ko=0,\\dots,so}`
    bound to oxygene :math:`i` and oxygene atoms :math:`o_{j,ks=0,\\dots,os}` bound to
    silicon :math:`j`, which are found using the verlet search algorithm.
    As an optimization, these values are not the atom ids, but the list pointers
    of the specific atoms in their corresponfing connection matrices :math:`si`
    and :math:`ox`. Due to chemical properties, the maximal possible silca atoms
    bound to one oxegene is :math:`so=2` and the maximal possible number of
    oxygene atoms bound to one silicon is :math:`os=4`.

    In the beginning, considering periodic boundary conditions and before
    removing any atoms, every atom should be saturated with partners. This means,
    That every silicon has a set of four oxygenes and every oxygene has two
    silicon bonds.

    With this connection matrix, bonds can be easily deleted and their number
    for specific atoms can be easily determined without needing another search
    call.

    Parameters
    ----------
    verlet : Verlet
        Verlet list object
    """

    def __init__(self, verlet):
        # Initialize
        osi = []
        sio = []
        pointer = []

        self._mol = verlet.get_mol()
        self._data = self._mol.get_data()
        self._dim = 3
        self._si_grid = 0.155
        self._y_grid = 0.073

        # Search for bounds
        bonds = verlet.find_parallel(None, ["Si", "O"], self._si_grid, 10e-2)

        # Fill atoms in list
        for i in range(self._mol.get_num()):
            atom_type = self._mol.get_type(i)
            if atom_type == "Si":
                pointer.append(len(sio))
                sio.append([i, []])
            elif atom_type == "O":
                pointer.append(len(osi))
                osi.append([i, []])

        # Transform to column
        sio = utils.column(sio)
        osi = utils.column(osi)

        # Add bonds to lists
        for i in range(len(bonds)):
            ids = pointer[bonds[i][0]]
            ido = [pointer[bonds[i][1][j]] for j in range(len(bonds[i][1]))]

            sio[1][ids] = ido
            for o in ido:
                osi[1][o].append(ids)

        # Make global
        self._sio = sio
        self._osi = osi
        self._pointer = pointer
        self._remove = []


    ###################
    # Private Methods #
    ###################
    def _unbind(self, silicon, oxygene):
        """Remove the bond between a silicon and an oxygene atom from the
        connection matrix.

        Parameters
        ----------
        silicon : integer
            Silica atom id
        oxygene : integer
            Oxygene atom id
        """
        # initialize
        sio = self._sio
        osi = self._osi
        pointer = self._pointer

        # Get pointer
        id_s = pointer[silicon]
        id_o = pointer[oxygene]

        # Remove bond
        sio[1][id_s].pop(sio[1][id_s].index(id_o))
        osi[1][id_o].pop(osi[1][id_o].index(id_s))

    def _geminal(self):
        """Get the list silicon that have gmeinal bonds - two binding sites.

        Returns
        -------
        geminal : list
            List of silicon with geminal bonds
        """
        # Get list of geminal silicon
        oxygene = [o[0] for o in self._osi[1] if len(o) == 1]
        silicon = list(set(oxygene))
        geminal = [self._sio[0][s] for s in silicon if oxygene.count(s) == 2]

        return geminal


    ############################
    # Public Methods - Editing #
    ############################
    def unlink(self, atoms):
        """Remove all bonds of a specified atom from the connection matrix.

        Parameters
        ----------
        atoms : list, integer
            List of atoms to be unlinked, can also be one atom id
        """
        # initialize
        sio = self._sio
        osi = self._osi
        pointer = self._pointer
        data = self._data

        # User input
        if not isinstance(atoms, list):
            atoms = [atoms]

        # Remove atoms
        for atom in atoms:
            # Get type
            atom_type = data[self._dim][atom]
            if atom_type == "Si":
                list_a = sio
                list_b = osi
            elif atom_type == "O":
                list_a = osi
                list_b = sio

            # Get indices
            id_a = pointer[atom]
            atoms_b = list_a[1][id_a]

            # Unlink
            for ab in atoms_b:
                list_b[1][ab].pop(list_b[1][ab].index(id_a))

            list_a[1][id_a] = []

    def remove(self, atoms):
        """Add an atom to a list of atoms to be deleted. This method is important,
        since by this call no bonds are removed.

        Parameters
        ----------
        atoms : list, integer
            List of atoms to be removed, can also be one atom id
        """
        # User input
        if not isinstance(atoms, list):
            atoms = [atoms]

        # Add to remove list
        self._remove.extend(atoms)

    def delete(self, is_list=False):
        """Delete all atoms without any bonds and the atoms in the remove list.

        Parameters
        ----------
        is_list : bool, optional
            True to return the atom list to be deleted, False to delte the atoms with no return value.

        Returns
        -------
        del : list
            List of atoms to be deleted
        """
        # Initialize
        sio = self._sio
        osi = self._osi

        # Search for empty bonds
        del_list = []
        lists = [sio, osi]

        for target in lists:
            for i in range(len(target[0])):
                if len(target[1][i]) == 0:
                    del_list.append(target[0][i])

        # Concatenate remove list
        del_list.extend(self._remove)
        del_list = list(set(del_list))

        # Return list or delete
        return del_list if is_list else self._mol.delete(del_list)


    ##############################
    # Public Methods - Structure #
    ##############################
    def attach(self):
        """The bonds in drilling direction are broken, since on these sites the
        reservoirs containing molecules are appended and binding sites
        are needed. After breaking the bonds, oxygenes are added to the unsaturated
        silicon on the outer surface.
        """
        # Initialize
        sio = self._sio
        osi = self._osi
        pointer = self._pointer
        data = self._data
        mol = self._mol
        si_grid = self._si_grid
        y_grid = self._y_grid

        drill = mol.get_drill()
        gap = mol.get_gap()
        box = mol.get_box_c()

        angle = 2*math.atan(math.sqrt(2))*180/math.pi

        # Atom adding helper function
        def add_oxy(si, o, dist, theta=0):
            # Process input
            if not isinstance(o,    list):
                o = [o]
            if not isinstance(theta, list):
                theta = [theta]

            # Run through oxygenes
            for i in range(len(o)):
                # Break bond
                self._unbind(sio[0][si], osi[0][o[i]])

                # Add atom
                mol.add("O", mol.pos(sio[0][si]), r=dist, theta=theta[i])

                # Add bond
                osi[0].append(mol.get_num()-1)
                osi[1].append([si])
                sio[1][si].append(len(osi[0])-1)
                pointer.append(len(osi[0])-1)

        # Adjustment helper function
        def adjust(ox, dist=None):
            # Process input
            if not isinstance(ox, list):
                ox = [ox]

            for o in ox:
                # Get positions
                posSi = mol.pos(sio[0][osi[1][o][0]])
                posO = mol.pos(osi[0][o])

                # Determine new position
                if dist is not None:
                    pos = posSi
                    pos[2] += dist
                else:
                    pos = posO
                    pos[1] = posSi[1]

                # Move atom
                mol.put(osi[0][o], pos)

        # Run through silicon molecules
        for i in range(len(sio[0])):
            # x-axis
            if drill == "x":
                if abs(data[2][sio[0][i]]-box[2]) < 10e-3:
                    for o in sio[1][i]:
                        if abs(data[2][osi[0][o]]-gap[2]) < 10e-3:
                            add_oxy(i, o, si_grid)
                            adjust(o, -si_grid)

            # y-axis
            elif drill == "y":
                # Silica at the front
                if abs(data[2][sio[0][i]]-box[2]+y_grid) < 10e-3:
                    for o in sio[1][i]:
                        if abs(data[2][osi[0][o]]-gap[2]) < 10e-3:
                            add_oxy(i, o, si_grid)
                            adjust(o, -si_grid)

                # Silica at the back
                if abs(data[2][sio[0][i]]-gap[2]) < 10e-3:
                    ox = []
                    theta = []
                    for o in sio[1][i]:
                        # Single oxygene
                        if abs(data[2][osi[0][o]]-box[2]+y_grid) < 10e-3:
                            add_oxy(i, o, -si_grid)
                            adjust(o, si_grid)
                            break

                        # Double oxygenes
                        if abs(data[2][osi[0][o]]-box[2]) < 10e-3:
                            # Find second silicon and get positions
                            si = osi[1][o][0] if not osi[1][o][0] == i else osi[1][o][1]
                            oP = data[0][osi[0][o]]
                            sP = data[0][sio[0][si]]

                            # Add to list
                            ox.append(o)

                            # Left, boundary left and right oxygene (this order)
                            theta.append(angle-180) if sP > oP or abs(sP -
                                                                      oP) > box[2]/2 else theta.append(180-angle)

                    add_oxy(i, ox, -si_grid, theta)
                    adjust(ox)

            # z-axis
            elif drill == "z":
                if abs(data[2][sio[0][i]]-gap[2]) < 10e-3:
                    for o in sio[1][i]:
                        if abs(data[2][osi[0][o]]-box[2]) < 10e-3:
                            add_oxy(i, o, -si_grid)

    def drill(self, focal, diam):
        """Drill through a pore through the silicon-ogygene-grid. This function
        does not delete any atoms but rather unlinks all bonds of atoms
        within the pore diameter.

        Parameters
        ----------
        focal : list
            Focal point of the grid
        diam : float
            Pore diamter
        """
        # Initialize
        data = self._data
        dim = self._dim
        sio = self._sio
        osi = self._osi

        # Delete silicon atoms
        lists = [sio, osi]

        for i in range(len(lists)):
            for target in lists[i][0]:
                length = math.sqrt(sum([(data[x][target]-focal[x])**2 for x in range(dim-1)]))

                if length < diam/2:
                    self.unlink(target)

    def prepare(self):
        """Prepare binding site in the pore. These are oxygene atoms with only
        one bond to a grid silicon and another loose end directed to the pore.

        Inspired by earlier work, the pore is then prepared in the steps

        1. Remove all unsaturated silicon, meaning all silicon with less than four bonds
        2. Remove all silicon with three binding sites

        Thus only silicon are left with a maximum of two binding sites.
        """
        # Initialize
        sio = self._sio
        osi = self._osi

        # Remove unsaturated silicon
        for i in range(len(sio[0])):
            if len(sio[1][i]) < 4:
                self.unlink(sio[0][i])

        # Remove silicon with three binding sites (2 times for safety)
        for i in range(2):
            temp_o = []
            for i in range(len(osi[0])):
                if len(osi[1][i]) == 1:
                    temp_o.extend(osi[1][i])

            temp_s = list(set(temp_o))

            for s in temp_s:
                if temp_o.count(s) > 2:
                    self.unlink(sio[0][s])

    def site(self):
        """Find and return the matrix :math:`\\boldsymbol{B}` of binding sites

        .. math::

            \\boldsymbol{B}=\\begin{bmatrix}
                o_0&s_{0,0}&t_0&g_0\\\\
                o_1&s_{1,0}&t_1&g_1\\\\
                \\vdots\\\\
                o_b&s_{b,0}&t_b&g_b
            \\end{bmatrix}

        with number of binding sites :math:`b`. The type :math:`t_{k=0,\\dots,b}`
        is 0 for binding sites inside the pore and 1 for sites on the outside.
        Geminal entries :math:`g_k` contain pointers to binding sites in
        the binding matrix :math:`\\boldsymbol{B}`.

        Returns
        -------
        site : list
            Binding site matrix
        """
        # Initialize
        osi = self._osi
        sio = self._sio
        data = self._data
        pointer = self._pointer
        geminal = self._geminal()
        box = self._mol.get_box_c()
        si_grid = self._si_grid
        site = []

        # Find binding sites - 0-O, 1-Si, 2-type, 3-geminal
        for i in range(len(osi[0])):
            if len(osi[1][i]) == 1:
                # Add indices
                entry = [osi[0][i], sio[0][osi[1][i][0]]]

                # Add type
                zCoord = data[2][entry[0]]
                if (abs(zCoord) < 10e-3 or abs(zCoord-box[2]) < 10e-3 or
                    abs(zCoord-si_grid) < 10e-3 or abs(zCoord-box[2]+si_grid) < 10e-3 or
                        abs(zCoord-0.107) < 10e-3 or abs(zCoord-box[2]+0.082) < 10e-3):
                    entry.append(1)
                else:
                    entry.append(0)

                # Add geminal
                entry.append(None)

                # Add entry to site
                site.append(entry)

        # Fill geminal
        temp_site = utils.column(site)
        for gem in geminal:
            indices = [i for i, x in enumerate(temp_site[1]) if x == gem]
            site[indices[0]][3] = indices[1]
            site[indices[1]][3] = indices[0]

        return site



################################################################################
# Pore Class                                                                   #
#                                                                              #
"""Extension of the molecule class for pores."""
################################################################################


import math
import copy
import random
import multiprocessing as mp

import porems.utils as utils

from porems.molecule import Molecule
from porems.verlet import Verlet
from porems.bonding import Bonding


class Pore(Molecule):
    """This class is an extension of the molecule class
    :class:`porems.molecule.Molecule` and the core class for creating
    silicon pores. Note that this class can therefore use all methods of the
    Molecule class.

    At the beginning a silicon-oxygene grid is genreated from the smallest
    possible grid unit by duplication. Using the verlet list
    :class:`porems.verlet.Verlet` and bonding classes
    :class:`porems.bonding.Bonding`, this silicon grid is prepared rapidly
    and binnding sites are provided.

    Here methods are provided for adding molecule groups to the
    surface of the generated pore. Furhermore a method is added to fill all
    unused binding sites with silanol and siloxan groups.

    All bound molecules are added to a global molecule list, in order to preserve
    individual molecule properties.

    Parameters
    ----------
    size : list, float
        Size of the silicon-oxygene-grid, can be a float, if same size in all dimensions
    diam : float
        Pore diameter
    drill : string
        Axis for the pore drilling
    res : float, optional
        Value to extend and translate the pore by on the drill side creating reservoirs
    vs : float, optional
        Verlet-list box size
    proxi_dist : float
        Minimal distance between two groups attatched to the surface
    is_pbc : bool, optional
        True if periodic boundary conditions are needed
    is_time : bool, optional
        True to print the used time for the single steps
    """
    def __init__(self, size, diam, drill, res=1, vs=0.4, proxi_dist=0.5, is_pbc=True, is_time=False):
        # Call super class
        super(Pore, self).__init__()

        # Initialize
        self._name = "pore"
        self._size = size if isinstance(size, list) else [size for s in range(self._dim)]
        self._diam = diam-{"x": 0.55, "y": 0.5, "z": 0.5}[drill]
        self._is_pbc = is_pbc
        self._is_time = is_time
        self._res = res
        self._vs = vs
        self._o_proxi = proxi_dist
        self._drill = drill

        # Define bond lengths
        self._si_grid = 0.155
        self._si_oh = 0.164
        self._oh = 0.098

        # Define silicon block information
        self._repeat = [0.506, 0.877, 1.240]
        self._gap = [0.126, 0.073, 0.155]
        self._num_rep = [round(self._size[i]/self._repeat[i]) for i in range(self._dim)]
        self._size = [self._repeat[i]*self._num_rep[i]+self._gap[i] for i in range(self._dim)]

        # Define charges
        self._q_si = 1.28
        self._q_o = -0.64
        self._q_oh = -0.74+0.42

        # Define lists
        self._site = []
        self._grid = []

        # Time management
        self._t_tot = {}

        # Properties
        self._props = {}
        self._is_props = False
        self._slx = 0

        # Create silicon pore
        t = utils.tic()
        self._block(0, self._foundation())             # Build block
        self._orientation()                            # Rotate drill axis
        self.translate(self._gap)                      # Translate gap
        self._center = self._focal()                   # Find focal point
        self._t_tot["Build"] = utils.toc(t, "Build   ", is_time)

        # Verlet and bonding
        t = utils.tic()
        self._verlet = Verlet(self, vs, is_pbc)        # Create verlet boxes
        self._bonding = Bonding(self._verlet)          # Create bond matrix
        self._t_tot["Matrix"] = utils.toc(t, "Matrix  ", is_time)

        # Prepare pore
        t = utils.tic()
        self._box = self.get_box()                     # Set box
        self._bonding.attach()                         # Preperare sites
        self._bonding.drill(self._center, self._diam)  # Drill pore
        self._bonding.prepare()                        # Prepare pore surface
        self._t_tot["Prepare"] = utils.toc(t, "Prepare ", is_time)

        # Find binding sites
        t = utils.tic()
        self.zero()
        self._box = self.get_box()                     # Reset box size
        self._bind()                                   # Create site array
        self._proxi()                                  # Find sites in proximity
        self._t_tot["Binding"] = utils.toc(t, "Binding ", is_time)

        # Calculate properties
        t = utils.tic()
        self.props()                                   # Recalculate properties
        self._diam = self._props["Diameter"]           # Set new diameter
        self._t_tot["Props"] = utils.toc(t, "Props   ", is_time)


    ############
    # Building #
    ############
    def _foundation(self):
        """In this method the smallest possible grid unit is created in a way,
        that duplication is simply done by copying the molecule without the need
        of deleting atoms in the end.

        Returns
        -------
        block : Molecule
            Smallest possible pore unit
        """
        # Initialize
        dist = self._si_grid
        angle = 2*math.atan(math.sqrt(2))*180/math.pi

        # Create pore
        h00 = Molecule()
        h00.add("Si", [0, 0, 0])
        h00.add("O", 0, r=dist, theta=angle, phi=60)
        h00.add("Si", 1, bond=[0, 1], r=dist)
        h00.add("O", 2, r=dist, theta=180,  phi=0)
        h00.add("Si", 3, bond=[2, 3], r=dist)
        h00.add("O", 4, r=dist, theta=angle, phi=300)
        h00.add("Si", 5, bond=[4, 5], r=dist)
        h00.add("O", 6, r=-dist, theta=angle, phi=60)
        h00.add("Si", 7, bond=[6, 7], r=dist)
        h00.add("O", 8, r=-dist, theta=180,  phi=0)
        h00.add("Si", 9, bond=[8, 9], r=dist)
        h00.add("O", 10, r=-dist, theta=angle, phi=300)

        h00.rotate("y", 90)
        h00.rotate("z", 90)
        h00.rotate("y", 180)
        h00.rotate("x", h00.angle(h00.bond(8, 6)[0], self._axis("y")))
        h00.rotate("x", -90)

        h01 = copy.deepcopy(h00)
        h01.add("O", 2, r=dist)
        h01.add("O", 6, r=dist)
        h01.add("O", 10, r=dist)

        h02 = copy.deepcopy(h00)
        h02.move(0, h01.pos(12))
        h02.translate([0, 0, dist])
        h02.delete([1, 2, 3, 4, 5, 6, 7])

        h03 = copy.deepcopy(h00)
        h03.move(0, h01.pos(14))
        h03.translate([0, 0, dist])
        h03.delete([2, 3, 4, 5, 6, 7, 8, 9, 10, 11])

        hexa = Molecule(inp=[h01, h02, h03])

        # Calculate xrepeat
        z = [hexa]
        for i in range(10):
            z.append(copy.deepcopy(hexa))

        z[1].rotate("x", 180)
        z[1].move(18, z[0].pos(18))
        z[1].translate([0, 0, dist*2])
        z[1].add("O", z[0].pos(18), r=dist)

        z[2].rotate("x", 180)
        z[2].move(0, z[0].pos(18))

        z[3].move(4, z[0].pos(15))

        z[4].rotate("x", 180)
        z[4].move(16, z[0].pos(18))
        z[4].delete([21, 19, 15, 20, 14, 12, 10, 2, 11, 1, 0])

        z[5].move(6, z[1].pos(16))
        z[5].delete([21, 19, 15, 20, 14, 12, 10, 2, 11, 1, 0])

        # Create repeat block
        block = Molecule(inp=z)
        block.overlap()

        # Determine number of needed blocks
        block.zero()

        # Delete overlapping atoms
        block.delete([4, 3, 2, 12, 15, 45, 46, 55, 57, 63, 26, 25, 24, 34, 37, 61])  # x-Axis
        block.delete([50, 48, 36, 40, 41]) # y-Axis
        block.delete([45, 20, 19, 21, 22, 23, 24, 25, 17, 18]) # z-Axis

        # Move to zero
        block.zero()

        return block

    def _block(self, dim, block):
        """Reculsively duplicate and translate a given molecule block in all
        given dimensions. The duplication stops if the next added block would
        create the pore longer than specified in the constructor.

        Parameters
        ----------
        dim : integer
            Repeat dimensions
        block : Molecule
            Molecule unit to be duplicated
        """
        if dim < self._dim:
            p = []

            for i in range(self._num_rep[dim]):
                temp = copy.deepcopy(block)
                vec = [(i+1)*self._repeat[dim] if j == dim else 0 for j in range(self._dim)]

                temp.translate(vec)

                p.append(temp)

            self._block(dim+1, Molecule(inp=p))
        else:
            self._append(block)

    def _orientation(self):
        """Rotate pore orientation, so that the specified drill axis becomes
        the z-axis.
        """
        # Initialize
        drill = self._drill
        gap = self._gap
        size = self._size

        # Rotate pore
        if drill == "x":
            self.rotate("y", 90)
        elif drill == "y":
            self.rotate("x", 90)

        # Set zero
        self.zero()

        # Update gap and size lists
        if drill == "x":
            self._gap = [gap[2], gap[1], gap[0]]
            self._size = [size[2], size[1], size[0]]
        elif drill == "y":
            self._gap = [gap[0], gap[2], gap[1]]
            self._size = [size[0], size[2], size[1]]


    #################
    # Binding sites #
    #################
    def _bind(self):
        """Create binding site matrix. This matrix :math:`\\boldsymbol{B}`
        is mostly generated by the bonding class :class:`poresms.bonding.Bonding`

        .. math::

            \\boldsymbol{B}=\\begin{bmatrix}
                o_0&s_0&p_0&u_0&t_0&g_0&m_0\\\\
                o_1&s_1&p_1&u_1&t_1&g_1&m_1\\\\
                \\vdots\\\\
                o_b&s_b&p_b&u_b&t_0&g_b&m_b\\\\
            \\end{bmatrix}

        with binding site number :math:`b` and entries

        0. Oxygen atom id :math:`o_{k=0,\\dots,b}`
        1. Silica atom id :math:`s_k`
        2. List of pointers :math:`p_k` to binding sites that are in proximity
        3. State :math:`u_k`: avilable - 0, used - 1, is in proximity - 2
        4. Type :math:`t_k`: inside the pore - 0, outsite the pore - 1
        5. Pointer :math:`g_k` to geminal binding site
        6. Molecule position :math:`m_k` in the global molecule list
        """
        # Initialize
        site = []

        # Get data
        osi = self._bonding.site()

        # Create bonding site list
        for os in osi:
            site.append([os[0], os[1], [], 0, os[2], os[3], -1])

        self._site.extend(site)

    def _proxi(self):
        """Using verlet lists, search for binding sites that are in proximity,
        by looking for oxygene atoms that are near each other.

        Fill the result in the binding site matrix in the proximity entry.
        """
        # Initialize
        site = self._site

        # Create local verlet list
        atoms = [o[0] for o in site]
        verlet = Verlet(self._temp(atoms), self._o_proxi, self._is_pbc)

        # Find oxygenes in proximity
        box_list = [i for i in range(len(verlet.get_box()[1]))]
        oo = verlet.find_parallel(box_list, ["O", "O"], 0.0, self._o_proxi)
        oo_col = utils.column(oo)

        # Append proxi list to sites
        for i in range(len(site)):
            os = site[i]
            o = oo[oo_col[0].index(i)][1] if i in oo_col[0] else []

            os[2].extend(o)

    def _random(self, site_type, rate, inp, counter, is_proxi=False):
        """Get a random binding site list for a given percentage of the
        available binding sites.

        If a site is available (state is 0), it is added to the output-list and
        the state entry for this site is set to 1.
        Then the states of all binding sites in the proximity list are set to 2.

        In case the random site is not available, this run is repeated.
        A break-counter prevents the occurance of an endless loop.

        The ``site_type`` determines wether the search is done on the inside of
        the pore **0** or on the outside surface **1**.

        Currently following rate types for input ``inp`` are available

        * **num** - For a specific number of binding sites
        * **percent** - For a percentage of the total number of binding sites on the selected surface
        * **molar** - For a molar value dependent on the surface area of the selected surface

        If ``is_proxi`` is set to True, then binding sites nearest to each other
        are determined. the return value then consists of lists containing the
        ids of both sites. Geminal pairs are not allowed, since creating the
        topology is too complex. This however might be added in the future.

        Parameters
        ----------
        site_type : integer
            Type of the sites inside-0, outside-1
        rate : float
            Rate of how many binding sites
        inp : string
            Input type of the rate
        counter : integer
            Number of attempts before breaking the randoming loop
        is_proxi : bool, optional
            True to search for site pairs in proximity

        Returns
        -------
        ran_list : list
            List of pointers to random binding sites
        """
        # Initialize
        site = copy.deepcopy(self._site)
        site_len = utils.column(site)[4].count(site_type)
        id_list = [i for i in range(len(site))]

        # Calculate random element number
        if inp == "num":
            ran_num = int(rate)
        elif inp == "percent":
            ran_num = site_len*rate
            ran_num /= 100
            ran_num = math.ceil(ran_num)
        elif inp == "molar":
            ran_num = rate*6.022/10*self._props["Surface"][site_type]
            ran_num = math.ceil(ran_num)

        # Get list of all closest partners for dual attachment
        if is_proxi:
            site_min = []
            for i in range(len(site)):
                length = [self._length(self._vector(site[i][0], site[j][0])) if not i ==
                          j and site[i][4] == site[j][4] else 100000 for j in range(len(site))]

                site_min.append([length.index(min(length)), min(length)])

        # Get random elements from site list
        ran_list = []
        count = 0
        while len(ran_list) < ran_num and count < counter:
            # Get random site index
            rand = random.choice(id_list)

            # Check binding site for dual attachment
            if is_proxi:
                is_app = (is_proxi and site[rand][4] == site_type and
                          site[rand][3] == 0 and site[site_min[rand][0]][3] == 0 and
                          site[rand][5] == None and site[site_min[rand][0]][5] == None)
            # Check binding site for singular attachment
            else:
                is_app = site[rand][4] == site_type and site[rand][3] == 0

            # Append to return list
            if is_app:
                # Reset counter
                count = 0

                # Append
                ran_list.append([rand, site_min[rand][0]] if is_proxi else rand)

                # Set site and proximity to used
                site = self._close([rand, site_min[rand][0]] if is_proxi else rand, site)
            else:
                # Add counter
                count += 1

        return ran_list

    def _close(self, sites, site_list=None):
        """This function closes the specified site by setting the used value
        to 1 and the used value of all sites in proximity to 2 in the given
        site list.

        Parameters
        ----------
        sites : integer, list
            identifiers of sites to be closed
        site_list : list, None, optional
            List for closing off binding sites, leave None for global list

        Returns
        -------
        site : list
            Processed site list
        """
        # Initialize
        site = self._site if site_list is None else site_list

        # Process input
        sites = [sites] if isinstance(sites, int) else sites

        # Close sites
        for x in sites:
            site[x][3] = 1
            for prox in site[x][2]:
                if not site[prox][3] == 1:
                    site[prox][3] = 2

        return site

    def _add_mol_list(self, mol_list):
        """Add given molecule list to global molecule list and add the position
        in this list to the corresponding binding site. The input has the format

        * [binding_site_idx, molecule]
        * [[binding_site_idx_1, molecule_1], [binding_site_idx_2, molecule_2], ...]

        Parameters
        ----------
        mol_list : list
            Molecules to be added to the global list
        """
        # Process input
        if not isinstance(mol_list[0],list):
            mol_list = [mol_list]

        # Add to molecule list
        for mol in mol_list:
            self._site[mol[1]][6] = len(self._mol_list)
            self._mol_list.append(mol[0])


    #######################
    # Molecule Attachment #
    #######################
    def attach(self, mol, si_o, orient, site_type, rate, inp="percent", sites=None, counter=1000,is_rotate=True):
        """Add a molecule to the pore binding site.

        The input ``si_o`` defines the silicon atom id and ``orient`` consists of
        two atom ids that determine which molecule bond should be oriented
        towords the center.

        The binding sites are chosen randomly using function :func:`_random`.
        Available input types are noted in the mentioned function.

        First the molecule is rotated around its directional vector randomly to
        simulate a state closer to reality. Second the molecule is rotated so that
        its directional vector and the binding site vector match. One final rotation
        is applied for molecules on the inside, so the directional vector
        points towards the center of the pore.

        The molecule is then moved in order to place its silicon atom on the silicon
        of the binding site. The oxygen is then placed ontop of the binding
        site oxygen, in order to keep the sites geometry.

        Finally the atoms of the binding sites are added to the list of atoms
        to be deleted at the end of the construction.

        These molecules are then added to the global molecule list.

        Parameters
        ----------
        mol : Molecule
            Molecule to be added
        si_o : list
            Silicon and oxygen id of the grid atoms of the molecule
        orient : list
            Two atom ids that determine which molecule bond should be oriented
            towords the center
        site_type : integer
            Type of the sites **inside-0**, **outside-1**
        rate : float
            Rate of how many binding sites
        inp : string, optional
            Input type of the rate
        sites : integer, list, optional
            Listindex of binding sites
        counter : integer, optional
            Number of attempts before breaking the randoming loop
        is_rotate : bool, optional
            True to randomly rotate molecules on the surface

        Returns
        -------
        mol_list : list
            List of Molecule objects and corresponding site ids
        remove_list : list
            List of atom ids to be deleted from the pore

        Examples
        --------
        >>> self.attach(TMS(), [0, 1], [1, 2], 0, 50)
        >>> self.attach(TMS(), [0, 1], [1, 2], 0, 50, inp="percent")
        >>> self.attach(TMS(), [0, 1], [1, 2], 1, 2.4, inp="molar")
        """
        # Stop time
        t = utils.tic()

        # Process user input
        sites = [sites] if isinstance(sites, int) else sites

        # Get random list or given list
        site_list = self._random(site_type, rate, inp, counter) if sites is None else sites

        # Molecule orientation vector
        vec_m = mol.bond(orient[0], orient[1])[0]

        # Center vector independet of z-axis
        center = self._center[:-1]

        # Add molcule
        mol_list = []
        remove_list = []
        for i in sorted(site_list, reverse=True):
            # Initialize
            site = self._site[i]

            # Copy molecule
            temp = copy.deepcopy(mol)

            # Randomly rotate molecule around orientational axis
            if is_rotate: temp.rotate(vec_m, random.randint(1, 180))

            # Set vector of silicon to oxygen atom of binding site
            vec_o = self._vector(site[1], site[0])

            # Adjust molecule vector to binding site vector
            temp.rotate(self._cross(vec_m, vec_o), self.angle(vec_m, vec_o))

            # Adjust molecule to face central axis on the inside
            if site[4] == 0:
                # Get center axis position perpendicular to binding site
                pos_c = center[:]+[self.pos(site[0])[self._dim-1]]

                # Set vector of binding site silicon atom to central position
                vec_c = self._vector(self.pos(site[1]), pos_c)

                # Rotate molecule towards center
                temp.rotate(self._cross(vec_c, vec_o), -self.angle(vec_c, vec_o))

            # Adjust orientation perpendicular to surface on the outside
            elif site[4] == 1:
                # Set vector towards or away from z-axis depending on which side
                vec_z = [x if self.pos(site[0])[2]>self._size[2]/2 else -1*x for x in self._axis("z")]

                # Rotate molecule perpendicular to surface
                temp.rotate(self._cross(vec_z, vec_o), -self.angle(vec_z, vec_o))

            # Move molecule to position of binding site originating from the oxygen atom
            temp.move(si_o[1], self.pos(site[0]))

            # Move molecule silicon atom to binding site position
            temp.put(si_o[0], self.pos(site[1]))

            # Check if geminal and set names
            if site[5] is not None:
                temp.set_name(temp.get_name()+"g")
                temp.set_short(temp.get_short()+"G")

            # Remove obsolete atoms
            remove_list.extend([site[0], site[1]])

            # Add to molecule list
            if sites == None:
                self._add_mol_list([temp, i])

            mol_list.append([temp, i])

            # Close binding sites
            self._close(i)

        # Finalize
        if not mol.get_name()=="sl":
            # Remove obsolete atoms
            self._bonding.remove(remove_list)

            # Initialize allocation
            key_name = mol.get_name()+"_"+str(site_type)

            # Add to time dict
            self._t_tot["Attach_"+key_name] = utils.toc(t, "Attach  ", self._is_time)

            # Add number of bonds to allocation
            if not key_name in self._props["Allocation"]:
                self._props["Allocation"][key_name] = 0
            self._props["Allocation"][key_name] += len(site_list)

        return mol_list, remove_list

    def special(self, mol, si_o, orient, num, symmetry):
        """Add a molecule in a specified orientation with a specific amount.
        Currently following symmetry orientations to each other are available

        * **random** - No symmetry, random gaussian placement
        * **point** - Point symmetry
        * **mirror** - Mirror symmetry

        Parameters
        ----------
        mol : Molecule
            Molecule to be added
        si_o : list
            Silicon and oxygen id of the grid atoms of the molecule
        orient : list
            Two atom ids that determine which molecule bond should be oriented
            towords the center
        num : integer
            Number of molecules to be added
        symmetry : string
            Molecule symmetry orientation to each other

        Examples
        --------
        >>> self.special(TMS(), [0, 1], [1, 2], 2)
        >>> self.special(TMS(), [0, 1], [1, 2], 2, symmetry="point")
        >>> self.special(TMS(), [0, 1], [1, 2], 3, symmetry="random")
        """
        # Process input
        if symmetry not in ["random", "point", "mirror"]:
            print("Symmetry type not supported...")
            return

        elif not symmetry == "random" and not num == 0:
            # Calculate placment points
            pos_list = []
            length = self._center[2]*2
            dist = length/num
            start = dist/2

            for i in range(num):
                if symmetry == "point":
                    coeff = -1 if i % 2 == 0 else 1
                elif symmetry == "mirror":
                    coeff = 1

                x = self._center[0]+coeff*self._diam/2
                y = self._center[1]
                z = start+dist*i

                pos_list.append([x, y, z])

            # Find nearest site
            site_min = [[0, None] for i in range(num)]
            for i in range(len(self._site)):
                length = [self._length(self._vector(self.pos(self._site[i][0]), pos)) for pos in pos_list]

                for j in range(num):
                    if site_min[j][1] is None or length[j] < site_min[j][1]:
                        site_min[j] = [i, length[j]]

            sites = [x[0] for x in site_min]

        else:
            sites = None

        mol_list, temp = self.attach(mol, si_o, orient, 0, num, inp="num", sites=sites, is_rotate=False)

        self._add_mol_list(mol_list)

    def _silanol(self, sites=None):
        """Convert the remaining binding sites to silanol groups using
        function :func:`attach`.

        Parameters
        ----------
        sites : list, None, optional
            Site list to go through

        Returns
        -------
        attach : dictionary
            List of Molecule objects and ids and List of atom ids to be deleted from the pore as a dictionary
        """
        # Initialize
        center = self._center[:-1]

        # Set siterange
        site_list = [i for i in range(len(self._site))] if sites is None else sites
        site_list = [x for x in site_list if not self._site[x][3] == 1]

        # Define temporary molecule object
        mol = Molecule()
        mol.add("Si",[0, 0, 0])
        mol.add("O", 0, r=self._si_oh)
        mol.add("H", 1, r=self._oh)
        mol.set_name("sl")
        mol.set_short("SL")
        mol.set_charge(self._q_si+self._q_oh)

        # Run attach method
        mol_list, remove_list = self.attach(mol, [0, 1], [0, 1], 0, len(site_list), inp="num", sites=site_list)

        if sites is None:
            self._add_mol_list(mol_list)
            self._bonding.remove(remove_list)
        else:
            return {"mol_list": mol_list, "remove_list": remove_list}

    def _silanol_parallel(self):
        """Parallelized function :func:`_silanol`.
        """
        # Initialize
        np = mp.cpu_count()

        # Split site list
        site_len = math.floor(len(self._site)/np)
        site_list = []
        for i in range(np):
            if i == np-1:
                site_list.append(list(range(site_len*i, len(self._site))))
            else:
                site_list.append(list(range(site_len*i, site_len*(i+1))))

        # Paralellize
        pool = mp.Pool(processes=np)
        results = pool.map_async(self._silanol, site_list)
        pool.close()

        # Extract results
        mol_list = []
        remove_list = []
        for result in results.get():
            mol_list.extend(result["mol_list"])
            remove_list.extend(result["remove_list"])

        # Add to mol_list and add site - molecule connection
        self._add_mol_list(mol_list)

        # Remove obsolete atoms
        self._bonding.remove(remove_list)

    def attach_dual(self, mol, si_o, orient, rate, inp="percent", counter=1000):
        """Add a molecule to two pore binding sites.

        The ``si_o`` input defines the silicon and oxygen atom pair in two seperate
        lists.

        First the molecule is rotated, so that the oxygens are ontop of the
        binding site oxygenes. Second the molecule is rotated so that the
        directional vector points to the center of the pore.

        The molecule is then moved, so that its oxygen atom is placed on the oxygen
        of the first binding site then translated, so the molecule is in the center
        of both sites.

        Finally the atoms of the binding sites are added to the list of atoms
        to be deleted at the end of the construction.

        Parameters
        ----------
        mol : Molecule
            Molecule to be added
        si_o : list
            List containing two silicon-oxygen vectors
        orient : list
            Molecule direction vector
        rate : float
            Rate of how many binding sites
        inp : string, optional
            Input type of the rate
        counter : integer, optional
            Number of attempts before breaking the randoming loop

        Examples
        --------
        >>> self.attach_dual(DualMol(), [[0, 2], [1, 3]], [4, 6], 80)
        >>> self.attach_dual(DualMol(), [[0, 2], [1, 3]], [4, 6], 80, inp="precent")
        >>> self.attach_dual(DualMol(), [[0, 2], [1, 3]], [4, 6], 10, inp="num")
        """
        # Get random list
        site_list = self._random(0, rate, inp, counter, is_proxi=True)

        # Molecule and center vector
        center = self._center[:-1]

        # Stop time
        t = utils.tic()

        # Add molcule
        for i in sorted(site_list, reverse=True):
            # Initialize
            sites = [self._site[i[0]], self._site[i[1]]]

            # Copy molecule
            temp = copy.deepcopy(mol)

            # Rotate towards binding site
            vec_mo = temp.bond(si_o[0][1], si_o[1][1])[0]   # Molecule oxygenes
            vec_so = self._vector(sites[0][0], sites[1][0]) # Binding site oxygenes
            vec_m = temp.bond(orient[0], orient[1])[0]      # Molecule orientation

            temp.rotate(vec_m, self.angle(vec_mo, vec_so))

            # Rotate towards central axis
            cent_o = [(self.pos(sites[0][0])[x]+self.pos(sites[1][0])[x])/2 for x in range(self._dim)]
            cent_p = center[:]+[cent_o[-1]]

            vec_m = temp.bond(orient[0], orient[1])[0]  # Molecule orientation
            vec_c = self._vector(cent_o, cent_p)        # Binding site center

            temp.rotate(self._cross(vec_m, vec_c), self.angle(vec_m, vec_c))

            # Put oxygenes ontop of each other
            temp.move(si_o[0][1], self.pos(sites[0][0]))

            # Move to center
            cent_m = [(temp.pos(si_o[0][1])[x]+temp.pos(si_o[1][1])[x])/2 for x in range(self._dim)]
            temp.translate(self._vector(cent_m, cent_o))

            # Move silicon atoms
            temp.put(si_o[0][0], self.pos(sites[0][1]))
            temp.put(si_o[1][0], self.pos(sites[1][1]))

            # Remove atoms
            self._bonding.remove([sites[0][0], sites[0][1], sites[1][0], sites[1][1]])

            # Add to molecule list
            self._mol_list.append(temp)

            # Close binding sites
            self._close(i)

        # Finalize
        key_name = mol.get_name()+"_0"

        # Add to time dict
        self._t_tot["Attach_"+key_name] = utils.toc(t, "Attach  ", self._is_time)

        # Add number of bonds to allocation
        if not key_name in self._props["Allocation"]:
            self._props["Allocation"][key_name] = 0
        self._props["Allocation"][key_name] += len(site_list)*2

    def siloxan(self, rate, inp="percent", counter=1000):
        """Add siloxan bridges to the pore by selecting two binding sites in
        proximity removing both oxygens and placing one at center of
        the two removed atoms.

        Parameters
        ----------
        rate : float
            Rate of binding sites to be editied
        inp : string, optional
            Rate input type
        counter : integer, optional
            Number of attempts before breaking the randoming loop

        Examples
        --------
        >>> self.siloxan(10)
        >>> self.siloxan(10, inp="precent")
        >>> self.siloxan(15, inp="num")
        >>> self.siloxan(2.4, inp="molar")
        """
        # Exit if rate is zero
        if rate == 0:
            return

        # Get random list
        site_list = self._random(0, rate, inp, counter, is_proxi=True)
        self._slx += len(site_list)

        # Molecule and center vector
        center = self._center[:-1]

        # Define temporary molecule object
        mol = Molecule()
        mol.set_name("slx")
        mol.set_short("SLX")
        mol.add("OM", [0, 0, 0])
        mol.add("O", 0, r=0.1)
        mol.add("O", 0, r=0.1)

        # Add molcule
        for i in sorted(site_list, reverse=True):
            # Initialize
            site = [self._site[i[0]], self._site[i[1]]]

            # Get center position
            cent_o = [(self.pos(site[0][0])[x]+self.pos(site[1][0])[x])/2 for x in range(self._dim)]
            cent_s = [(self.pos(site[0][1])[x]+self.pos(site[1][1])[x])/2 for x in range(self._dim)]

            # Create molecule
            temp = copy.deepcopy(mol)
            temp.put(0, cent_o)
            temp.translate(self._vector(cent_s, cent_o))
            temp.put(1, self.pos(site[0][0]))
            temp.put(2, self.pos(site[1][0]))

            # Rotate towards central axis
            cent_p = center[:]+[cent_o[-1]]

            vec_m = self._vector(cent_o, temp.pos(0))
            vec_c = self._vector(cent_o, cent_p)

            temp.rotate(self._cross(vec_m, vec_c), self.angle(vec_m, vec_c))

            # Put oxygenes ontop of each other
            temp.move(1, self.pos(site[0][0]))

            # Remove temporary oxygen atoms
            temp.delete([1, 2])

            # Remove atoms
            self._bonding.remove([site[0][0], site[1][0]])

            # Add to molecule list
            self._mol_list.append(temp)

            # Close binding sites
            self._close(i)


    ###############
    # Final Edits #
    ###############
    def _connect(self):
        """Concatenate two geminal molecules to one molecule.
        """
        # Initialize
        site = self._site
        mol_list = self._mol_list
        gemi = []

        # Check for geminal sites
        for i in range(len(site)):
            if site[i][5] is not None:
                gem = [site[i][6], site[site[i][5]][6]]
                gem_r = [site[site[i][5]][6], site[i][6]]

                if not gem_r in gemi:
                    gemi.append(gem)

        # Append atoms
        pop = []
        for g in gemi:
            id_a = 0 if mol_list[g[1]].get_name() == "slg" else 1
            id_b = abs(id_a-1)

            mol_a = mol_list[g[id_a]]
            mol_b = mol_list[g[id_b]]

            num = mol_b.get_num()

            for i in range(num):
                if not mol_b.get_type(i) == "Si":
                    mol_a.add(mol_b.get_type(i), mol_b.pos(i))

            pop.append(g[id_b])

            # Add charge
            mol_list[g[id_a]].set_charge(mol_list[g[id_a]].get_charge()+self._q_oh)

        # Delete from mol_list list
        for d in sorted(pop, reverse=True):
            mol_list.pop(d)

    def _objectify(self, data=None):
        """Move all remaining grid silicon and oxygen atoms to individual molecules
        for writing the structure file and add these molecules to the global
        molecule list.

        Parameters
        ----------
        data_range : list, None, optional
            Datarange to process

        Returns
        -------
        mol_list : list
            Molecule list to be added to globale molecule list
        """
        # Define temporary molecule object
        temp_mol = Molecule()
        data_range = [0,len(self._data[0])] if data is None else data

        # Create OM and SI mols
        mol_list = []
        for i in range(data_range[0], data_range[1]):
            temp = copy.deepcopy(temp_mol)

            # Check if oxygene
            is_oxy = self.get_type(i) == "O"

            # Set temporary vars
            atom = "OM" if is_oxy else "SI"
            name = "om" if is_oxy else "si"
            short = "OM" if is_oxy else "SI"
            charge = self._q_o if is_oxy else self._q_si

            # Testing
            if self.get_type(i) == "R":
                atom = "R"
                name = "test"

            # Creat unique object
            temp.add(atom, self.pos(i))
            temp.set_name(name)
            temp.set_short(short)
            temp.set_charge(charge)

            # Add to molecule list
            if data is None:
                self._mol_list.append(temp)
            else:
                mol_list.append(temp)

        # Add mols to molecule list
        if data is None:
            self._mol_list.pop(0)
        else:
            return mol_list

    def _objectify_parallel(self):
        """Parallelized function :func:`_objectify`.
        """
        # Initialize
        np = mp.cpu_count()

        # Split site list
        data_len = math.floor(len(self._data[0])/np)
        data_list = []
        for i in range(np):
            if i == np-1:
                data_list.append([data_len*i, len(self._data[0])])
            else:
                data_list.append([data_len*i, data_len*(i+1)])

        # Paralellize
        pool = mp.Pool(processes=np)
        results = pool.map_async(self._objectify, data_list)
        pool.close()

        # Extract results
        mol_list = []
        for result in results.get():
            mol_list.extend(result)

        # Add to molecule list
        for mol in mol_list:
            self._mol_list.append(mol)

        # Remove pore from molecule list
        self._mol_list.pop(0)

    def _position(self):
        """Tranlate the pore into position, and consider the periodic repeat gap.
        Hereby the coordinates are moved by halve the gap distance and box box is
        extended by the other halve. Additionally, to prevent molecule overlap and
        perodic movement, the drill side is extended and translated by a fixed value,
        thus creating reservoirs.
        """
        # Get vector to coordinate zero
        vec = Molecule(inp=[mol for mol in self._mol_list if mol.get_name() in ["si", "om"]]).zero()

        # Position pore
        for mol in self._mol_list:
            mol.translate([vec[i]+self._gap[i]/2 for i in range(self._dim)])
            mol.translate([self._res if i == 2 else 0 for i in range(self._dim)])

        # Extend box
        coord = Molecule(inp=[mol for mol in self._mol_list if mol.get_name() in ["si", "om", "ox"]]).get_box()

        self._box = [coord[i] + self._gap[i]/2 for i in range(self._dim)]

        self._box[2] += self._res

    def _overlap(self):
        """Method for checking of silicon or oxygene atoms are overlapping
        using verlet lists. Duplicate atoms will be printed.
        """
        # Initialize
        temp = Molecule(inp=self._mol_list)

        # Create verlet
        verlet = Verlet(temp, self._vs, True)

        # Check silicon and oxygene
        si = verlet.find_parallel(None, ["Si", "Si"], 0, 10e-3)
        oo = verlet.find_parallel(None, ["O", "O"], 0, 10e-3)

        # Output
        for s in si:
            if len(s[1]) > 0:
                print(s)

        for o in oo:
            if len(o[1]) > 0:
                print(o)

    def _excess(self, is_rand=True):
        """Adjust grid atoms charges to compensate for excess charges on the pore.

        Parameters
        ----------
        is_rand : bool, optional
            True to randomly distribute rounding error charge on block oxygen atoms
        """
        # Initialize
        mols = self._mol_list

        # Get excess charge
        charge = 0.0
        num_o = 0
        num_si = 0
        num_g = 0
        round_o = 6
        for mol in mols:
            charge += mol.get_charge()

            if mol.get_name() == "om":
                num_o += 1
            elif mol.get_name() == "si":
                num_si += 1
            else:
                num_g += 1

        # Calculate adjustment
        avg = charge/(num_o/2+num_si+num_g)
        q_o = round(self._q_o - avg/2, round_o)
        q_si = round(self._q_si-avg,  round_o)

        # Distribute charge
        for mol in mols:
            if mol.get_name() == "om":
                mol.set_charge(q_o)
            elif mol.get_name() == "si":
                mol.set_charge(q_si)
            else:
                mol.set_charge(round(mol.get_charge()-self._q_si+q_si, round_o))

        # Calculate rounding error
        charge_err = sum([mol.get_charge() for mol in mols])
        num_oxy = 100 if num_o > 100 else num_o
        avg_err = charge_err/num_oxy
        q_ox = round(q_o-avg_err, round_o)

        # Randomly distribute balancing oxygens
        if is_rand:
            # Get list of oxygenes
            oxy = {}
            for i in range(len(mols)):
                if mols[i].get_name() == "om":
                    oxy[i] = mols[i]

            # Get random oxygenes and set name and charge
            oxy_r = random.sample([x for x in oxy], num_oxy)
            for o in oxy_r:
                oxy[o].set_name("ox")
                oxy[o].set_short("OX")
                oxy[o].set_charge(q_ox)

            # Push oxy after om
            min_o = min([x for x in oxy])
            max_o = max([x for x in oxy])
            mol_list = []
            mol_list_ox = []
            for i in range(min_o):
                mol_list.append(mols[i])
            for i in range(min_o, max_o+1):
                if i not in oxy_r:
                    mol_list.append(mols[i])
                else:
                    mol_list_ox.append(mols[i])
            mol_list += mol_list_ox
            for i in range(max_o+1, len(mols)):
                mol_list.append(mols[i])

            self._mol_list = mol_list

        # Make global
        self._q = {"om": q_o, "ox": q_ox, "si": q_si}

        self.set_charge(sum([mol.get_charge() for mol in self._mol_list]))

    def _sort(self):
        """Sort molecules in order to prevent multiple molecule appearances in
        the structure file.
        """
        # Add molecules to a dictionary
        mol_dict = {}
        for mol in self._mol_list:
            if mol.get_name() not in mol_dict:
                mol_dict[mol.get_name()] = []

            mol_dict[mol.get_name()].append(mol)

        # Create new molecule list
        mol_list = []
        priority = ["si","om","ox","slx","sl","slg"]
        for mol_name in priority:
            if mol_name in mol_dict:
                mol_list += mol_dict[mol_name]

        for mol_name in mol_dict:
            if mol_name not in priority:
                mol_list += mol_dict[mol_name]

        # Replace global list
        self._mol_list = mol_list


    ###########
    # Analyze #
    ###########
    def props(self):
        """This function calculates all necessary properties of the system and
        retruns a dictionary. Note that most of the properties can only be
        calculated after the pore is finalized in function :func:`finalize`.

        The output dictionary contains following properties which will be
        explained further down

        * **Allocation** - Dictionary of all molecule allocations in molar and percent
        * **Charge** - Excess charge
        * **Diameter** - Generated pore diameter
        * **Dimension** - Generated Dimensions
        * **Generation_Time** - Dictionary of all time steps
        * **Hydroxylation** - Hydroxylation dictionary inside and outside of the pore
        * **Roughness** - Surface roughness
        * **Silanol_Geminal** - Dictionary containing the number of silanol and geminal silanol molecules inside and outside the pore
        * **Surface** - Dictionary of inner and outer pore surface
        * **System_Size** - Total system size including reservoirs
        * **Volume** - Inner pore volume

        The **roughness** is calculated as the standard deviation of the peaks and
        valles of a surface.

        In the case of a pore one can visualize pulling it appart, creating a
        flat surface out of the interior. The roughness is thus determined by
        calculating the standard deviation of the binding site silicon peaks and
        valleys.

        The only difference in the pore is therefore the definition of the axis,
        which is going to be the centeral axis. The mean value :math:`\\bar r`
        of the silicon distances :math:`r_i` of silicon :math:`i` towards the
        pore center, is calculated by

        .. math::

            \\bar r=\\frac1n\\sum_{i=1}^nr_i

        with the number of silicon atoms :math:`n`. This mean value is used in
        the sqareroot roughness calculation

        .. math::

            R_q = \\sqrt{\\frac1n\\sum_{i=1}^n\\|r_i-\\bar r\\|^2}.


        The **diameter** is the mean value :math:`\\bar r` of the silicon distances
        :math:`r_i` of binding site silicon :math:`i` towards the pore center

        .. math::

            d=2\\bar r=\\frac2n\\sum_{i=1}^nr_i.

        The **surface** inside the pore is calculated by

        .. math::

            A_\\text{inside}=2\\pi r\\cdot l_\\text{drill},

        with opening radius :math:`r` and length of the drilling axis
        :math:`l_\\text{drill}`, the total outside surface of the pore system by

        .. math::

            A_\\text{outside}=2\\cdot\\left(A_\\text{side}-\\pi r^2\\right)

        with block surface on the drilling side :math:`A_\\text{side}`,
        and the **volume** by

        .. math::

            V_\\text{pore}=\\pi r^2\\cdot l_\\text{drill}.

        Using these surfaces, the **allocation** rates can be determined by
        counting the number of used sites and free sites from the binding
        site list. Thus it is possible to determine the maximal possible allocation
        :math:`c_\\text{tot}` also called hydroxylation, the rate for the
        molecule modification :math:`c_\\text{mod}` and the rate for the silanol
        modification :math:`c_\\text{oh}`. For better overview, the relative
        allocation for the molecule modification :math:`c_\\text{rel}` has been
        calculated

        .. math::

            \\begin{array}{cccc}
            c_\\text{tot} = \\dfrac{s_\\text{tot}}{A},&
            c_\\text{mod} = \\dfrac{s_\\text{mod}}{A},&
            c_\\text{oh}  = \\dfrac{s_\\text{oh} }{A},&
            c_\\text{rel} = \\dfrac{s_\\text{mod}}{s_\\text{tot}}
            \\end{array}

        with the corresponding surface :math:`A` for either inside or outside
        the pore, the total number of binding sites :math:`s_\\text{tot}`,
        number of sites used for the modification :math:`s_\\text{mod}`
        and number of sites used for the silanol modification :math:`s_\\text{oh}`.
        The resulting units are

        .. math::

            [c_{i}]=\\frac{\\text{Number of molecules}}{nm^2}
            =\\frac{1}{N_A}\\frac{mol}{nm^2}
            =\\frac{10}{6.022}\\frac{\\mu mol}{m^2}.

        Returns
        -------
        props : dictionary
            Properties of the pore
        """
        # Initialize
        props = self._props

        # Calculate generation properties
        if not self._is_props:
            # Run through binding site silicon atoms and calculate distances
            r_list = []
            for site in self._site:
                if site[4] == 0:
                    si_id = site[1]
                    center = [self._center[0], self._center[1], self._data[2][si_id]]
                    r_list.append(self._length(self._vector(self.pos(si_id), center)))

            # Calculate mean
            r_bar = sum(r_list)/len(r_list) if len(r_list)>0 else 0

            # Set dimension order
            dim_order = {"x": [2, 1, 0], "y": [0, 2, 1], "z": [0, 1, 2]}

            # Calculate properties
            props["Roughness"] = math.sqrt(sum([(ri-r_bar)**2 for ri in r_list])/len(r_list)) if r_bar>0 else 0
            props["Diameter"] = 2*r_bar if r_bar>0 else 0
            props["Surface"] = {0: self._size[2]*2*math.pi*(props["Diameter"]/2),
                                1: 2*(self._size[0]*self._size[1]-math.pi*(props["Diameter"]/2)**2)}
            props["Volume"] = self._size[2]*math.pi*(props["Diameter"]/2)**2
            props["Dimension"] = [self._size[dim_order[self._drill][i]] for i in range(3)]
            props["Allocation"] = {}

        # Calculate finished properties
        else:
            # Molar conversion
            molar = 10/6.022

            # Count number of groups
            num_i = sum([1 for x in self._site if x[4] == 0 and x[5] is None])
            num_o = sum([1 for x in self._site if x[4] == 1 and x[5] is None])
            num_i_g = sum([1 for x in self._site if x[4] == 0 and x[5] is not None])/2
            num_o_g = sum([1 for x in self._site if x[4] == 1 and x[5] is not None])/2

            num_tot = [num_i+num_i_g-2*self._slx,num_o+num_o_g]

            # Calculate hydroxylation
            hydro = [num_tot[0]/props["Surface"][0], num_tot[1]/props["Surface"][1]]

            # Calculate allocation
            num_oh = [num_i+num_i_g, num_o+num_o_g]
            mols = [mol for mol in props["Allocation"]]

            for mol in mols:
                site_type = int(mol[-1])
                mod = props["Allocation"][mol]/props["Surface"][site_type]
                rel = props["Allocation"][mol]/num_tot[site_type]

                # Add molar
                props["Allocation"][mol+"_mod"] = [mod, mod*molar]
                props["Allocation"][mol+"_rel"] = [rel, rel*molar]

                # Count oh
                num_oh[site_type] -= props["Allocation"][mol]

            # OH allocation
            oh = [num_oh[0]/props["Surface"][0], num_oh[1]/props["Surface"][1]]
            props["Allocation"]["OH"] = {"in": [oh[0], oh[0]*molar], "out": [oh[1], oh[1]*molar]}

            # Siloxan allocation
            slx = self._slx/props["Surface"][0]
            props["Allocation"]["SLX"] = [slx, slx*molar]

            # Calculate properties
            props["Charge"] = self._charge
            props["System_Size"] = self._size
            props["Generation_Time"] = self._t_tot
            props["Hydroxylation"] = {"in": [hydro[0],hydro[0]*molar], "out": [hydro[1],hydro[1]*molar]}
            props["Silanol_Geminal"] = {"in": num_i, "out": num_o, "in_g": num_i_g, "out_g": num_o_g}

        # Return properties
        return props


    #########################
    # Public Methods - Edit #
    #########################
    def finalize(self, is_rand=True):
        """Finalize pore by adding siloxan bridges, filling empty bond with
        silanol groups, connecting geminal molecules, removing marked atoms,
        converting the grid to individual molecules and finally moving the pore
        into position.

        Parameters
        ----------
        is_rand : bool, optional
            True to randomly distribute rounding error charge on block oxygen atoms
        """
        # Fill empty sites with silanol
        t = utils.tic()
        self._silanol_parallel()
        self._t_tot["Silanol"] = utils.toc(t, "Silanol ", self._is_time)

        # Connect geminal molecules into one
        t = utils.tic()
        self._connect()
        self._t_tot["Connect"] = utils.toc(t, "Connect ", self._is_time)

        # Delete marked atoms
        t = utils.tic()
        self._bonding.delete()
        self._t_tot["Delete"] = utils.toc(t, "Delete  ", self._is_time)

        # Move all silicon and oxygenes into unique mols
        t = utils.tic()
        self._objectify_parallel()
        self._t_tot["Objects"] = utils.toc(t, "Objects ", self._is_time)

        # Distribute excess charge
        t = utils.tic()
        self._excess(is_rand)
        self._t_tot["Excess"] = utils.toc(t, "Excess  ", self._is_time)

        # Sort molecules
        t = utils.tic()
        self._sort()
        self._t_tot["Sort"] = utils.toc(t, "Sort    ", self._is_time)

        # Position the pore considering pbc
        t = utils.tic()
        self._position()
        self._t_tot["Position"] = utils.toc(t, "Position", self._is_time)

        # Check for silicon or oxygen atoms overlapping
        t = utils.tic()
        self._overlap()
        self._t_tot["Overlap"] = utils.toc(t, "Overlap ", self._is_time)

        # Allow properties calculation
        self._is_props = True

        if self._is_time:
            print("----------------------------")
            print("Total   - runtime = "+"%6.3f" % sum([self._t_tot[x] for x in self._t_tot])+" s")
            print()

        # Print finalization note
        print("Pore is finalized. Attachement functions should not be called anymore.")


    ##################
    # Getter Methods #
    ##################
    def get_time(self):
        """Return the total creation time.

        Returns
        -------
        time : float
            Used pore creation time
        """
        return self._t_tot

    def get_size(self):
        """Return the pore size.

        Returns
        -------
        size : list
            Pore dimension
        """
        return self._size

    def get_diam(self):
        """Return the pore diameter.

        Returns
        -------
        diam : float
            Pore diameter
        """
        return self._diam

    def get_reservoir(self):
        """Return the reservoir length.

        Returns
        -------
        res : float
            Reservoir length
        """
        return self._res

    def get_q(self):
        """Return the adjusted grid charges.

        Returns
        -------
        q : dictionary
            Dictionary of grid molecules and their charges
        """
        return self._q

    def get_drill(self):
        """Return the drilling axis.

        Returns
        -------
        drill : string
            Drilling axis
        """
        return self._drill

    def get_gap(self):
        """Return the gap vector.

        Returns
        -------
        gap : list
            Gap vector
        """
        return self._gap

    def get_center(self):
        """Return the pore focal point.

        Returns
        -------
        center : list
            Pore focal point
        """
        return self._center

    def get_props(self):
        """Return the final pore properties.

        Returns
        -------
        props : dictionary
            Dictionary containing properties of the generated pore
        """
        return self._props



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="porems",
    version="0.1.0",
    author="Hamzeh Kraus",
    author_email="kraus@itt.uni-stuttgart.de",
    description="Pore Generator for Molecular Simulations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ajax23/PoreMS",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    include_package_data=True,
)



################################################################################
# Store Class                                                                  #
#                                                                              #
"""Contains function for creating simululation files."""
################################################################################


import os
import shutil

from decimal import Decimal

import porems.utils as utils

from porems.molecule import Molecule
from porems.pore import Pore


class Store:
    """This class converts a molecule object to a structure file of different
    possible formats. Currently only **PDB** and **GRO** are fully supported.

    For pore objects two additional methods are available for generating the
    main topology file and a topology file containing grid molecule
    parameters and charges. Currently only the gromacs format is supported.

    Furthermore there is an automized rountine for generating topologies
    with Antechamber where job-file for this tool are created.

    Parameters
    ----------
    molecule : Molecule
        Molecule object
    link : string, optional
        Folder link for output
    """
    def __init__(self, molecule, link="./"):
        # Get molecule properties
        self._dim = 3
        self._link = link if link[-1] == "/" else link+"/"
        self._mol = molecule
        self._mols = molecule.get_mol_list()
        self._name = molecule.get_name()
        self._short = molecule.get_short()
        self._box = Molecule(inp=self._mols).get_box() if molecule.get_box_c() == None else molecule.get_box_c()

        # Create output folder
        utils.mkdirp(link)


    ################################
    # Public Methods - Antechamber #
    ################################
    def job(self, is_master=False):
        """Create job file to run with Antechamber. A shell file and a tleap file
        are created with all necesarry commands to create a topology from a pdb file.
        For the conversion to gromacs file format the python package **ParmED**
        needs to be installed.

        Parameters
        ----------
        is_master : bool, optional
            True if the jobfile call should be added to a master run file
            (pracitcall for running multiple topology genrations)
        """
        # Initialize
        name = self._name.lower()
        short = self._short
        link = self._link

        # Template directory
        package_dir = os.path.split(__file__)[0]+"/"

        # Job file
        file_in = package_dir+"templates/antechamber.job"
        file_out = link+name+".job"
        shutil.copy(file_in, file_out)

        utils.replace(file_out, "MOLNAME", name)

        # Tleap file
        file_in = package_dir+"templates/antechamber.tleap"
        file_out = link+name+".tleap"
        shutil.copy(file_in, file_out)

        utils.replace(file_out, "MOLSHORTLOWER", name.lower())
        utils.replace(file_out, "MOLSHORT", short)
        utils.replace(file_out, "MOLNAME", name)

        # Add to master run
        if is_master:
            fileMaster = open(link+"run.job", "a")
            fileMaster.write("cd "+name+" #\n")
            fileMaster.write("sh "+name+".job #\n")
            fileMaster.write("cd .. #\n")
            fileMaster.write("echo \"Finished "+name+"...\"\n")
            fileMaster.write("#\n")
            fileMaster.close()


    ##############################
    # Public Methods - Structure #
    ##############################
    def obj(self, name=None):
        """Save the molecule object using pickle.

        Parameters
        ----------
        name : None, string, optional
            Filename
        """
        # Initialize
        link = self._link
        link += self._name+".obj" if name is None else name

        # Save object
        utils.save(self._mol, link)

    def pdb(self, name=None):
        """Generate the structure file for the defined molecule in the **PDB** format.

        Parameters
        ----------
        name : None, string, optional
            Filename
        """
        # Initialize
        dim = self._dim
        mols = self._mols
        link = self._link
        link += self._name+".pdb" if name is None else name

        # Open file
        with open(link, "w") as file_out:
            # Atoms
            num_a = 1
            num_m = 1
            for mol_id in range(len(mols)):
                data = mols[mol_id].get_data()
                short = mols[mol_id].get_short()
                ids = {x: 1 for x in data[dim]}
                for i in range(len(data[0])):
                    atom = data[dim][i]
                    out_string = "HETATM"                        #  1- 6 (6)    Record name
                    out_string += "%5i" % num_a                  #  7-11 (5)    Atom serial number
                    out_string += " "                            # 12    (1)    -
                    out_string += "%4s" % (atom+str(ids[atom]))  # 13-16 (4)    Atom name
                    # 17    (1)    Alternate location indicator
                    out_string += " "
                    out_string += "%3s" % short                  # 18-20 (3)    Residue name
                    out_string += " "                            # 21    (1)    -
                    out_string += "%1s" % "A"                    # 22    (1)    Chain identifier
                    # 23-26 (4)    Residue sequence number
                    out_string += "%4i" % (num_m)
                    # 27    (1)    Code for insertion of residues
                    out_string += " "
                    out_string += "   "                          # 28-30 (3)    -

                    for j in range(dim):                         # 31-54 (3*8)  Corrdinates
                        out_string += "%8.3f" % (data[j][i]*10)

                    out_string += "%6.2f" % 1                    # 55-60 (6)    Occupancy
                    out_string += "%6.2f" % 0                    # 61-66 (6)    Temperatur factor
                    out_string += "           "                  # 67-77 (11)   -
                    out_string += "%2s" % atom                   # 78-79 (2)    Element symbol
                    out_string += "  "                           # 80-81 (2)    Charge on the atom

                    file_out.write(out_string+"\n")

                    ids[atom] += 1
                    num_a += 1

                num_m = num_m+1 if num_m < 9999 else 1

            # End statement
            file_out.write("TER\nEND\n")

    def gro(self, name=None):
        """Generate the structure file for the defined molecule in the **GRO** format.

        Parameters
        ----------
        name : None, string, optional
            Filename
        """
        # Initialize
        dim = self._dim
        box = self._box
        mols = self._mols
        link = self._link
        link += self._name+".gro" if name is None else name

        # Open file
        with open(link, "w") as file_out:
            # Set title
            file_out.write("Molecule generated by the PoreMS package\n")

            # Number of atoms
            file_out.write("%5i" % sum([len(x.get_data()[0]) for x in mols])+"\n")

            # Atoms
            num_a = 1
            num_m = 1
            for mol_id in range(len(mols)):
                data = mols[mol_id].get_data()
                short = mols[mol_id].get_short()
                ids = {x: 1 for x in data[dim]}
                for i in range(len(data[0])):
                    atom = data[dim][i]
                    atom_name = atom+str(ids[atom])

                    out_string = "%5i" % num_m  # 1- 5 (5)    Residue number
                    out_string += "%-5s" % short  # 6-10 (5)    Residue short name
                    out_string += "%5s" % atom_name  # 11-15 (5)    Atom name
                    out_string += "%5i" % num_a      # 16-20 (5)    Atom number
                    for j in range(dim):            # 21-44 (3*8)  Coordinates
                        out_string += "%8.3f" % data[j][i]

                    file_out.write(out_string+"\n")

                    num_a = num_a+1 if num_a < 99999 else 0
                    ids[atom] += 1

                num_m = num_m+1 if num_m < 99999 else 0

            # Box
            out_string = ""
            for i in range(dim):
                out_string += "%.3f" % box[i]
                out_string += " " if i < dim-1 else "\n"

            file_out.write(out_string)

    def xyz(self, name=None):
        """Generate the structure file for the defined molecule in the xyz
        format for running qm-simulations.

        Parameters
        ----------
        name : string, None, optional
            Filename
        """
        # Initialize
        mols = self._mols
        link = self._link
        link += self._name+".xyz" if name is None else name

        # Open output file and set title
        with open(link, "w") as file_out:
            file_out.write(str(sum([len(x.get_data()[0]) for x in mols]))+"\n"+"Energy = \n")

            # Atoms
            for mol_id in range(len(mols)):
                data = mols[mol_id].get_data()
                for i in range(len(data[0])):
                    out_string = "%-2s" % data[-1][i]  # 1- 2 (2)     Atom name

                    for j in range(self._dim):       # 3-41 (3*13)  Coordinates
                        out_string += "%13.7f" % (data[j][i]*10)

                    file_out.write(out_string+"\n")


    ############
    # Topology #
    ############
    def top(self, name=None):
        """Store the topology file for a pore. A top file is created containing
        the itp-include for all molecules and the count of the different groups
        of the pore. If parameter ``link`` is given, parameter ``name`` will be
        ignored.

        Parameters
        ----------
        name : None, string, optional
            Filename
        """
        # Initialize
        mols = self._mols
        link = self._link
        link += self._name+".top" if name is None else name

        # Get unique molecules
        unique_mols = []
        for mol in mols:
            if not mol.get_name() in unique_mols:
                unique_mols.append(mol.get_name())

        # Open file
        with open(link, "w") as file_out:
            # Store header
            file_out.write("[ defaults ]\n")
            file_out.write("; nbfunc        comb-rule       gen-pairs       fudgeLJ fudgeQQ\n")
            file_out.write("1               2               yes             0.5     0.833333\n")
            file_out.write("\n")
            file_out.write("#include \"grid.itp\"\n")

            for mol_name in unique_mols:
                if mol_name not in ["si", "om", "ox", "sl", "slg", "slx"]:
                    file_out.write("#include \""+mol_name+".itp\"\n")

            file_out.write("\n")
            file_out.write("[ system ]\n")
            file_out.write("A pore used in molecular simulation plus reservoir\n")
            file_out.write("\n")
            file_out.write("[ molecules ]\n")

            # Atoms
            counter = 1
            for i in range(1, len(mols)):
                if mols[i].get_name() == mols[i-1].get_name():
                    counter += 1
                else:
                    file_out.write(mols[i-1].get_short()+" "+str(counter)+"\n")
                    counter = 1

            file_out.write(mols[-1].get_short()+" "+str(counter)+"\n")

    def grid(self, name=None):
        """Store the grid.itp file containing the necesarry parameters and
        charges of the grid molecules.

        Parameters
        ----------
        name : None, string, optional
            Filename
        """
        # Initialize
        link = self._link
        link += "grid.itp" if name is None else name

        # Calculate excess charge
        charges = self._mol.get_q()

        # Copy grid file
        utils.copy(os.path.split(__file__)[0]+"/templates/grid.itp", link)

        # Replace charges
        utils.replace(link, "CHARGEOX", "%8.6f" % charges["ox"])
        utils.replace(link, "CHARGEO", "%8.6f" % charges["om"])
        utils.replace(link, "CHARGESI", "%8.6f" % charges["si"])



################################################################################
# Essential Pack                                                               #
#                                                                              #
"""This file contains essential molecule classes and serves as an exemplary
molecule generation process."""
################################################################################


from porems.molecule import Molecule


class Alkane(Molecule):
    """This class extends the molecule class :class:`porems.molecule.Molecule`.
    Here linear alkane molecules can be easily constructed with the only
    input of the carbon atom number. Hydrogenes are then automatically added.

    Parameters
    ----------
    length : integer
        Number of carbon atoms
    name : string, None, optional
        Molecule name
    short : string, None, optional
        Molecule short name
    is_h : bool, optional
        True if hydrogenes are needed
    """
    def __init__(self, length, name=None, short=None, is_h=True):
        # Call super class
        super(Alkane, self).__init__()

        # Define molecule names
        if name is not None:
            self.set_name(name)
        if short is not None:
            self.set_short(short)

        # Define bond lengths and angles
        b = {"cc": 0.153, "ch": 0.109}
        a = {"ccc": 30.00, "cch": 109.47}

        # Add carbons
        self.add("C", [0, 0, 0])

        angle = a["ccc"]
        for i in range(length-1):
            angle *= -1
            self.add("C", self.get_num()-1, r=b["cc"], theta=angle)

        # Add hydrogenes
        if is_h:
            if length > 1:
                # Boundaries
                end = self.get_num()-1

                for i in range(3):
                    self.add("H", 0, bond=[0, 1], r=b["ch"], theta=a["cch"], phi=60+120*i)
                    self.add("H", end, bond=[end, end-1], r=b["ch"], theta=a["cch"], phi=60+120*i)

                # Interior
                angle = -90
                for i in range(1, length-1):
                    angle *= -1
                    self.add("H", i, r=b["ch"], theta=angle, phi=a["cch"])
                    self.add("H", i, r=b["ch"], theta=angle, phi=-a["cch"])

            # Methane
            else:
                self.add("H", 0, r=b["ch"])
                for i in range(3):
                    self.add("H", 0, r=b["ch"], theta=a["cch"], phi=i*120)

        # Move to zero
        self.zero()


class Alcohol(Molecule):
    """This class extends the molecule class :class:`porems.molecule.Molecule`.
    Here linear alcohol molecules can be easily constructed with the only
    input of the carbon atom number. Hydrogenes and the hydroxy group are then
    automatically added.

    Parameters
    ----------
    length : integer
        Number of carbon atoms
    name : string, None, optional
        Molecule name
    short : string, None, optional
        Molecule short name
    is_h : bool, optional
        True if hydrogenes are needed
    """
    def __init__(self, length, name=None, short=None, is_h=True):
        # Call super class
        super(Alcohol, self).__init__()

        # Define molecule names
        if name is not None:
            self.set_name(name)
        if short is not None:
            self.set_short(short)

        # Define bond lengths and angles
        b = {"cc": 0.153, "ch": 0.109, "co": 0.143, "oh": 0.098}
        a = {"ccc": 30.00, "cch": 109.47, "occ": 30.00, "coh": 109.47}

        # Add carbons
        self.add("C", [0, 0, 0])

        angle = a["ccc"]
        for i in range(length-1):
            angle *= -1
            self.add("C", self.get_num()-1, r=b["cc"], theta=angle)

        # Add hydroxy
        self.add("O", self.get_num()-1, r=b["co"], theta=-angle)
        self.add("H", self.get_num()-1, r=b["oh"], theta=angle)

        # Add hydrogenes
        if is_h:
            if length > 1:
                # Boundary
                for i in range(3):
                    self.add("H", 0, bond=[0, 1], r=b["ch"], theta=a["cch"], phi=120*i)

                # Interior
                angle = -90
                for i in range(1, length):
                    angle *= -1
                    self.add("H", i, r=b["ch"], theta=angle, phi=a["cch"])
                    self.add("H", i, r=b["ch"], theta=angle, phi=-a["cch"])

            # Methanol
            else:
                for i in range(3):
                    self.add("H", 0, r=b["ch"], theta=a["cch"], phi=i*120)

        # Move to zero
        self.zero()


class Ketone(Molecule):
    """This class extends the molecule class :class:`porems.molecule.Molecule`.
    Here linear ketone molecules can be easily constructed with the only
    input of the carbon atom number and oxygene position.
    Hydrogenes are then automatically added.

    Parameters
    ----------
    length : integer
        Number of carbon atoms
    pos : integer
        Position of the oxygene atom
    name : string, None, optional
        Molecule name
    short : string, None, optional
        Molecule short name
    is_h : bool
        True if hydrogenes are needed
    """
    def __init__(self, length, pos, name=None, short=None, is_h=True):
        # Call super class
        super(Ketone, self).__init__()

        # Check input
        if length < 3:
            print("Specified length is too small for ketones ...")
            return

        # Define molecule names
        if name is not None:
            self.set_name(name)
        if short is not None:
            self.set_short(short)

        # Define bond lengths and angles
        b = {"cc": 0.153, "ch": 0.109, "co": 0.123}
        a = {"ccc": 30.00, "cch": 109.47}

        # Add carbons
        self.add("C", [0, 0, 0])

        angle = a["ccc"]
        for i in range(length-1):
            angle *= -1
            self.add("C", self.get_num()-1, r=b["cc"], theta=angle)

        # Add oxygene
        angle = -90 if pos % 2 == 0 else 90
        self.add("O", pos-1, r=b["co"], theta=angle)

        # Add hydrogenes
        if is_h:
            # Boundaries
            end = self.get_num()-1

            for i in range(3):
                self.add("H", 0, bond=[0,  1], r=b["ch"], theta=a["cch"], phi=60+120*i)
                self.add("H", end-1, bond=[end-1, end-2], r=b["ch"], theta=a["cch"], phi=60+120*i)

            # Interior
            angle = -90
            for i in range(1, length-1):
                angle *= -1
                if not i == pos-1:
                    self.add("H", i, r=b["ch"], theta=angle, phi=a["cch"])
                    self.add("H", i, r=b["ch"], theta=angle, phi=-a["cch"])

        # Move to zero
        self.zero()


class TMS(Molecule):
    """This class extends the molecule class :class:`porems.molecule.Molecule`
    and defines trimethylsilyl (TMS) bound to a silicon-grid.

    Parameters
    ----------
    name : string, optional
        Molecule name
    short : string, optional
        Molecule short name
    charge : float, optional
        Excess charge of the entire grid molecule
    compress : float, optional
        Compress molecule sidechain angles
    is_si : bool, optional
        True if the terminus should be a lone silicon, False for a CH3 group
    """
    def __init__(self, name="tms", short="TMS", charge=1.070428, compress=0, is_si=True):
        # Call super class
        super(TMS, self).__init__()

        # Define molecule names
        self.set_name(name)
        self.set_short(short)

        # Check silicon
        si = "Si" if is_si else "Ci"

        # Define bond lengths and angles
        b = {"sio": 0.155, "sic": 0.186, "ch": 0.109}
        a = {"ccc": 30.00, "cch": 109.47, "siosi": 126.12}

        # Build silyl chain
        self.add(si, [0, 0, 0])
        self.add("O", 0, r=b["sio"])
        self.add(si, 1, bond=[1, 0], r=b["sio"], theta=a["siosi"])

        # Add methyl
        for i in range(3):
            self.add("C", 2, bond=[2, 1], r=b["sic"], theta=a["cch"]+compress, phi=60+120*i)

        # Add hydrogenes
        for i in range(3, 5+1):
            for j in range(3):
                self.add("H", i, bond=[i, 2], r=b["ch"], theta=a["cch"]+compress, phi=60+120*j)

        # If not silicon ending
        for i in range(3):
            if not is_si:
                self.add("H", 0, bond=[0, 1], r=b["ch"], theta=a["cch"], phi=60+120*i)

        # Set charge
        self.set_charge(charge)
